# type: ignore

import re
from types import SimpleNamespace

from prompt_toolkit.document import Document
import pytest

import mycli.sqlcompleter
from mycli.sqlcompleter import Fuzziness, SQLCompleter


def collect_matches(
    orig_text: str,
    collection: list[str],
    *,
    start_only: bool = False,
    fuzzy: bool = True,
    casing: str | None = None,
    text_before_cursor: str = '',
) -> list[tuple[str, int]]:
    completer = SQLCompleter()
    return list(
        completer.find_matches(
            orig_text,
            collection,
            start_only=start_only,
            fuzzy=fuzzy,
            casing=casing,
            text_before_cursor=text_before_cursor,
        )
    )


def make_completer(**kwargs) -> SQLCompleter:
    comp = SQLCompleter(**kwargs)
    comp.keywords = list(comp.keywords)
    comp.functions = list(comp.functions)
    return comp


@pytest.mark.parametrize(
    ('item', 'expected'),
    [
        ('users', '`users`'),
        ('`already`', '`already`'),
        ('*', '*'),
    ],
)
def test_maybe_quote_identifier(item: str, expected: str) -> None:
    completer = SQLCompleter()
    assert completer.maybe_quote_identifier(item) == expected


def test_quote_collection_if_needed_quotes_when_text_starts_with_backtick() -> None:
    completer = SQLCompleter()
    quoted = completer.quote_collection_if_needed('`us', ['users', '*'], '')

    assert quoted == ['`users`', '*']


def test_quote_collection_if_needed_quotes_when_cursor_is_inside_backticks() -> None:
    completer = SQLCompleter()
    quoted = completer.quote_collection_if_needed('us', ['users', '`uuid`'], 'select `us')

    assert quoted == ['`users`', '`uuid`']


def test_quote_collection_if_needed_leaves_collection_unchanged_when_not_quoted() -> None:
    collection = ['users', '*']
    completer = SQLCompleter()
    quoted = completer.quote_collection_if_needed('us', collection, 'select us')

    assert quoted is collection


@pytest.mark.parametrize(
    ('text_parts', 'item_parts', 'expected'),
    [
        (['us', 'de', 'fu'], ['user', 'defined', 'function'], True),
        (['us', 'fu'], ['user', 'defined', 'function'], True),
        (['us', 'zz'], ['user', 'defined', 'function'], False),
        ([], ['user', 'defined', 'function'], True),
        (['us'], [], False),
    ],
)
def test_word_parts_match(
    text_parts: list[str],
    item_parts: list[str],
    expected: bool,
) -> None:
    completer = SQLCompleter()
    assert completer.word_parts_match(text_parts, item_parts) is expected


@pytest.mark.parametrize(
    ('item', 'pattern', 'under_words_text', 'case_words_text', 'expected'),
    [
        ('foo_select_bar', re.compile('(s.{0,3}?e.{0,3}?l)'), ['sel'], ['sel'], Fuzziness.REGEX),
        ('user_defined_function', re.compile('(z.{0,3}?z)'), ['us', 'de', 'fu'], ['us_de_fu'], Fuzziness.UNDER_WORDS),
        ('TimeZoneTransitionType', re.compile('(Ti.{0,3}?Zx)'), ['TiZoTrTy'], ['Ti', 'Zo', 'Tr', 'Ty'], Fuzziness.CAMEL_CASE),
        ('orders', re.compile('(z.{0,3}?z)'), ['zz'], ['zz'], None),
    ],
)
def test_find_fuzzy_match(
    item: str,
    pattern: re.Pattern[str],
    under_words_text: list[str],
    case_words_text: list[str],
    expected: int | None,
) -> None:
    completer = SQLCompleter()
    assert completer.find_fuzzy_match(item, pattern, under_words_text, case_words_text) == expected


def test_find_fuzzy_matches_collects_item_level_matches(monkeypatch) -> None:
    monkeypatch.setattr(
        SQLCompleter,
        'find_fuzzy_match',
        lambda self, item, pattern, under_words_text, case_words_text: {
            'orders': Fuzziness.REGEX,
            'order_items': Fuzziness.UNDER_WORDS,
            'other': None,
        }[item],
    )
    monkeypatch.setattr(mycli.sqlcompleter.rapidfuzz.process, 'extract', lambda *args, **kwargs: [])
    completer = SQLCompleter()
    matches = completer.find_fuzzy_matches('OrIt', 'orit', ['orders', 'order_items', 'other'])

    assert matches == [
        ('orders', Fuzziness.REGEX),
        ('order_items', Fuzziness.UNDER_WORDS),
    ]


def test_find_fuzzy_matches_skips_rapidfuzz_for_short_text(monkeypatch) -> None:
    monkeypatch.setattr(SQLCompleter, 'find_fuzzy_match', lambda *args, **kwargs: None)

    def fail_extract(*args, **kwargs):
        raise AssertionError('rapidfuzz should not be called')

    monkeypatch.setattr(mycli.sqlcompleter.rapidfuzz.process, 'extract', fail_extract)
    completer = SQLCompleter()
    matches = completer.find_fuzzy_matches('sel', 'sel', ['SELECT'])

    assert matches == []


def test_find_fuzzy_matches_appends_rapidfuzz_results_and_skips_duplicates(monkeypatch) -> None:
    monkeypatch.setattr(
        SQLCompleter,
        'find_fuzzy_match',
        lambda self, item, pattern, under_words_text, case_words_text: Fuzziness.REGEX if item == 'alphabet' else None,
    )
    monkeypatch.setattr(
        mycli.sqlcompleter.rapidfuzz.process,
        'extract',
        lambda *args, **kwargs: [('abc', 99, 0), ('alphabet', 95, 1), ('alphanumeric', 90, 2)],
    )
    completer = SQLCompleter()
    matches = completer.find_fuzzy_matches('alpahet', 'alpahet', ['abc', 'alphabet', 'alphanumeric'])

    assert matches == [
        ('alphabet', Fuzziness.REGEX),
        ('alphanumeric', Fuzziness.RAPIDFUZZ),
    ]


@pytest.mark.parametrize('existing_fuzziness', [Fuzziness.PERFECT, Fuzziness.CAMEL_CASE, Fuzziness.RAPIDFUZZ])
def test_find_fuzzy_matches_skips_rapidfuzz_duplicates_for_remaining_fuzziness_types(
    monkeypatch,
    existing_fuzziness: Fuzziness,
) -> None:
    monkeypatch.setattr(
        SQLCompleter,
        'find_fuzzy_match',
        lambda self, item, pattern, under_words_text, case_words_text: existing_fuzziness if item == 'alphabet' else None,
    )
    monkeypatch.setattr(
        mycli.sqlcompleter.rapidfuzz.process,
        'extract',
        lambda *args, **kwargs: [('alphabet', 95, 0)],
    )
    completer = SQLCompleter()

    matches = completer.find_fuzzy_matches('alpahet', 'alpahet', ['alphabet'])

    assert matches == [('alphabet', existing_fuzziness)]


@pytest.mark.parametrize(
    ('text', 'collection', 'start_only', 'expected'),
    [
        ('ord', ['orders', 'user_orders'], True, [('orders', Fuzziness.PERFECT)]),
        ('name', ['table_name', 'name_table'], False, [('table_name', Fuzziness.PERFECT), ('name_table', Fuzziness.PERFECT)]),
        ('', ['orders', 'users'], True, [('orders', Fuzziness.PERFECT), ('users', Fuzziness.PERFECT)]),
    ],
)
def test_find_perfect_matches(
    text: str,
    collection: list[str],
    start_only: bool,
    expected: list[tuple[str, int]],
) -> None:
    completer = SQLCompleter()
    assert completer.find_perfect_matches(text, collection, start_only) == expected


@pytest.mark.parametrize(
    ('casing', 'last', 'expected'),
    [
        (None, 'Sel', None),
        ('upper', 'sel', 'upper'),
        ('lower', 'SEL', 'lower'),
        ('auto', 'sel', 'lower'),
        ('auto', 'SEl', 'lower'),
        ('auto', 'SEL', 'upper'),
        ('auto', '', 'upper'),
    ],
)
def test_resolve_casing(casing: str | None, last: str, expected: str | None) -> None:
    completer = SQLCompleter()
    assert completer.resolve_casing(casing, last) == expected


@pytest.mark.parametrize(
    ('completions', 'casing', 'expected'),
    [
        ([('Select', Fuzziness.REGEX)], None, [('Select', Fuzziness.REGEX)]),
        ([('Select', Fuzziness.REGEX)], 'upper', [('SELECT', Fuzziness.REGEX)]),
        ([('Select', Fuzziness.REGEX)], 'lower', [('select', Fuzziness.REGEX)]),
        (
            [('Select', Fuzziness.REGEX), ('From', Fuzziness.PERFECT)],
            'upper',
            [('SELECT', Fuzziness.REGEX), ('FROM', Fuzziness.PERFECT)],
        ),
    ],
)
def test_apply_casing(
    completions: list[tuple[str, int]],
    casing: str | None,
    expected: list[tuple[str, int]],
) -> None:
    completer = SQLCompleter()
    assert list(completer.apply_casing(completions, casing)) == expected


def test_find_matches_uses_last_word_for_prefix_matching() -> None:
    matches = collect_matches(
        'select ord',
        ['orders', 'user_orders'],
        start_only=True,
        fuzzy=False,
    )

    assert matches == [('orders', Fuzziness.PERFECT)]


def test_find_matches_supports_substring_matching() -> None:
    matches = collect_matches(
        'name',
        ['table_name', 'name_table'],
        start_only=False,
        fuzzy=False,
    )

    assert matches == [
        ('table_name', Fuzziness.PERFECT),
        ('name_table', Fuzziness.PERFECT),
    ]


def test_find_matches_quotes_identifiers_when_text_starts_with_backtick() -> None:
    matches = collect_matches('`us', ['users'])

    assert matches == [('`users`', Fuzziness.REGEX)]


def test_find_matches_quotes_identifiers_when_cursor_is_inside_backticks() -> None:
    matches = collect_matches(
        'uu',
        ['users', '`uuid`'],
        text_before_cursor='select `uu',
    )

    assert matches == [('`uuid`', Fuzziness.REGEX)]


def test_find_matches_preserves_asterisk_inside_backticks() -> None:
    matches = collect_matches(
        '*',
        ['*'],
        text_before_cursor='select `*',
    )

    assert matches == [('*', Fuzziness.REGEX)]


def test_find_matches_finds_regex_matches() -> None:
    matches = collect_matches('sel', ['SELECT', 'foo_select_bar'])

    assert matches == [
        ('SELECT', Fuzziness.REGEX),
        ('foo_select_bar', Fuzziness.REGEX),
    ]


def test_find_matches_finds_under_word_matches() -> None:
    matches = collect_matches('us_de_fu', ['user_defined_function'])

    assert matches == [('user_defined_function', Fuzziness.UNDER_WORDS)]


def test_find_matches_finds_camel_case_matches(monkeypatch) -> None:
    monkeypatch.setattr(mycli.sqlcompleter.rapidfuzz.process, 'extract', lambda *args, **kwargs: [])

    matches = collect_matches('TiZoTrTy', ['TimeZoneTransitionType'])

    assert matches == [('TimeZoneTransitionType', Fuzziness.CAMEL_CASE)]


def test_find_matches_finds_rapidfuzz_matches() -> None:
    matches = collect_matches('sleect', ['SELECT'])

    assert matches == [('SELECT', Fuzziness.RAPIDFUZZ)]


def test_find_matches_skips_rapidfuzz_for_short_text(monkeypatch) -> None:
    def fail_extract(*args, **kwargs):
        raise AssertionError('rapidfuzz should not be called')

    monkeypatch.setattr(mycli.sqlcompleter.rapidfuzz.process, 'extract', fail_extract)

    matches = collect_matches('sel', ['SELECT'])

    assert matches == [('SELECT', Fuzziness.REGEX)]


def test_find_matches_filters_short_rapidfuzz_candidates(monkeypatch) -> None:
    monkeypatch.setattr(
        mycli.sqlcompleter.rapidfuzz.process,
        'extract',
        lambda *args, **kwargs: [('abc', 99, 0), ('alphabet', 95, 1)],
    )

    matches = collect_matches('alpahet', ['abc', 'alphabet'])

    assert matches == [('alphabet', Fuzziness.RAPIDFUZZ)]


@pytest.mark.parametrize(
    ('orig_text', 'collection', 'casing', 'expected'),
    [
        ('sel', ['SELECT'], 'auto', [('select', Fuzziness.REGEX)]),
        ('SEL', ['select'], 'auto', [('SELECT', Fuzziness.REGEX)]),
        ('sel', ['select'], 'upper', [('SELECT', Fuzziness.REGEX)]),
        ('SEL', ['SELECT'], 'lower', [('select', Fuzziness.REGEX)]),
    ],
)
def test_find_matches_applies_casing(
    orig_text: str,
    collection: list[str],
    casing: str,
    expected: list[tuple[str, int]],
) -> None:
    matches = collect_matches(orig_text, collection, casing=casing)

    assert matches == expected


def test_init_invalid_keyword_casing_defaults_to_auto() -> None:
    completer = SQLCompleter(keyword_casing='invalid')

    assert completer.keyword_casing == 'auto'


def test_extend_metadata_helpers_and_logging(caplog) -> None:
    completer = make_completer()
    completer.set_dbname('missing')

    completer.extend_keywords(['ZZZ'])
    assert 'ZZZ' in completer.keywords
    assert 'ZZZ' in completer.all_completions

    completer.extend_keywords(['ONLY_THIS'], replace=True)
    assert completer.keywords == ['ONLY_THIS']
    assert 'ONLY_THIS' in completer.all_completions

    completer.extend_show_items([('FULL TABLES',), ('STATUS',)])
    completer.extend_change_items([('MASTER TO',)])
    completer.extend_users([('app_user',)])
    assert completer.show_items == ['FULL TABLES', 'STATUS']
    assert 'MASTER TO' in completer.change_items
    assert 'app_user' in completer.users

    completer.extend_schemata(None)
    assert '' not in completer.dbmetadata['tables']

    with caplog.at_level('ERROR', logger='mycli.sqlcompleter'):
        completer.extend_relations([('orders',)], kind='tables')
    assert "listed in unrecognized schema 'missing'" in caplog.text

    completer.extend_schemata('test')
    completer.set_dbname('test')
    completer.extend_relations([('select',)], kind='tables')

    caplog.clear()
    with caplog.at_level('ERROR', logger='mycli.sqlcompleter'):
        completer.extend_columns([('missing', 'id'), ('select', 'from')], kind='tables')
    assert "relname 'missing' was not found in db 'test'" in caplog.text
    assert completer.dbmetadata['tables']['test']['`select`'] == ['*', '`from`']

    completer.set_dbname('enumdb')
    completer.extend_enum_values([('order status', 'select', ['pending'])])
    assert completer.dbmetadata['enum_values']['enumdb']['`order status`']['`select`'] == ['pending']


def test_extend_functions_procedures_character_sets_and_collations() -> None:
    completer = make_completer()
    completer.extend_schemata('test')
    completer.set_dbname('test')

    completer.extend_functions(['BUILTIN_X'], builtin=True)
    assert 'BUILTIN_X' in completer.functions

    def broken_functions():
        raise RuntimeError('boom')
        yield ('ignored', 'ignored')

    completer.extend_functions(broken_functions())
    completer.extend_functions(iter([('quoted func', 'meta')]))
    assert '`quoted func`' in completer.dbmetadata['functions']['test']

    completer.extend_procedures(iter([(), (None,), ('proc_demo',)]))
    assert 'proc_demo' in completer.dbmetadata['procedures']['test']

    completer.extend_character_sets(iter([(), (None,), ('utf8mb4',)]))
    completer.extend_collations(iter([(), (None,), ('utf8mb4_unicode_ci',)]))
    assert completer.character_sets == ['utf8mb4']
    assert completer.collations == ['utf8mb4_unicode_ci']


def test_extend_procedures_initializes_schema_metadata_when_missing() -> None:
    completer = make_completer()
    completer.set_dbname('procdb')

    completer.extend_procedures(iter([('proc_demo',)]))

    assert completer.dbmetadata['procedures']['procdb']['proc_demo'] is None


def test_get_completions_drop_unique_columns(monkeypatch) -> None:
    completer = make_completer()
    completer.extend_schemata('test')
    completer.set_dbname('test')
    completer.dbmetadata['tables']['test'] = {
        't1': ['*', 'id', 'name'],
        't2': ['*', 'id', 'email'],
    }

    monkeypatch.setattr(
        mycli.sqlcompleter,
        'suggest_type',
        lambda text, before: [{'type': 'column', 'tables': [(None, 't1', None), (None, 't2', None)], 'drop_unique': True}],
    )

    result = [c.text for c in completer.get_completions(Document(text='SELECT ', cursor_position=7), None)]

    assert result == ['id']


@pytest.mark.parametrize(
    ('suggestion', 'setup', 'text', 'expected'),
    [
        ({'type': 'procedure', 'schema': 'test'}, lambda c, m: c.extend_procedures(iter([('proc_demo',)])), 'CALL pro', 'proc_demo'),
        ({'type': 'show'}, lambda c, m: c.extend_show_items([('TABLE STATUS',)]), 'SHOW tab', 'table status'),
        ({'type': 'change'}, lambda c, m: c.extend_change_items([('MASTER TO',)]), 'CHANGE ma', 'MASTER TO'),
        ({'type': 'user'}, lambda c, m: c.extend_users([('app_user',)]), 'GRANT app', 'app_user'),
        (
            {'type': 'favoritequery'},
            lambda c, m: m.setattr(
                mycli.sqlcompleter.FavoriteQueries, 'instance', SimpleNamespace(list=lambda: ['daily_report']), raising=False
            ),
            '\\f dai',
            'daily_report',
        ),
        ({'type': 'table_format'}, lambda c, m: None, 'fmt c', 'csv'),
    ],
)
def test_get_completions_branch_specific_suggestions(monkeypatch, suggestion, setup, text, expected) -> None:
    completer = make_completer(supported_formats=('csv', 'tsv'))
    completer.extend_schemata('test')
    completer.set_dbname('test')
    setup(completer, monkeypatch)
    monkeypatch.setattr(mycli.sqlcompleter, 'suggest_type', lambda full_text, before: [suggestion])

    result = [c.text for c in completer.get_completions(Document(text=text, cursor_position=len(text)), None)]

    assert expected in result


def test_get_completions_llm_branch_with_and_without_current_word(monkeypatch) -> None:
    tokens_seen: list[list[str]] = []

    def fake_get_completions(tokens: list[str]) -> list[str]:
        tokens_seen.append(tokens)
        return ['chat', 'explain']

    monkeypatch.setattr(mycli.sqlcompleter, 'suggest_type', lambda full_text, before: [{'type': 'llm'}])
    monkeypatch.setattr(mycli.sqlcompleter.llm, 'get_completions', fake_get_completions)

    completer = make_completer()

    blank_word = [c.text for c in completer.get_completions(Document(text='\\llm ', cursor_position=5), None)]
    partial_text = '\\llm ask ch'
    partial_word = [c.text for c in completer.get_completions(Document(text=partial_text, cursor_position=len(partial_text)), None)]

    assert tokens_seen == [[], ['ask']]
    assert 'chat' in blank_word
    assert 'chat' in partial_word
    assert 'explain' in blank_word
    assert 'explain' not in partial_word


def test_find_files_populate_scoped_cols_and_enum_helpers(monkeypatch) -> None:
    completer = make_completer()
    completer.extend_schemata('test')
    completer.set_dbname('test')
    completer.dbmetadata['tables']['test']['`select`'] = ['id']
    completer.dbmetadata['views']['test']['orders_view'] = ['view_id']
    completer.extend_enum_values([('orders', 'status', ['pending', 'shipped'])])

    monkeypatch.setattr(mycli.sqlcompleter, 'parse_path', lambda word: ('/tmp', 'fi', 0))
    monkeypatch.setattr(mycli.sqlcompleter, 'suggest_path', lambda word: ['file.sql', 'folder/'])
    monkeypatch.setattr(mycli.sqlcompleter, 'complete_path', lambda name, last_path: name if name == 'file.sql' else None)

    assert list(completer.find_files('./fi')) == [('file.sql', Fuzziness.PERFECT)]
    assert completer.populate_scoped_cols([(None, 'select', None), (None, 'orders_view', None), (None, 'missing', None)]) == [
        'id',
        'view_id',
    ]
    assert completer.populate_enum_values([(None, 'orders', 'o')], 'status', parent='other') == []
    assert completer.populate_enum_values([(None, 'orders', 'o')], 'status', parent='o') == ['pending', 'shipped']
    assert completer._quote_sql_string("O'Reilly") == "'O''Reilly'"


@pytest.mark.parametrize(
    ('name', 'expected'),
    [
        ('`quoted`', 'quoted'),
        ('plain', 'plain'),
        (None, ''),
    ],
)
def test_strip_backticks(name: str | None, expected: str) -> None:
    assert SQLCompleter._strip_backticks(name) == expected


@pytest.mark.parametrize(
    ('parent', 'schema', 'relname', 'alias', 'expected'),
    [
        ('o', None, 'orders', 'o', True),
        ('orders', None, 'orders', None, True),
        ('test.orders', 'test', 'orders', None, True),
        ('other', 'test', 'orders', 'o', False),
    ],
)
def test_matches_parent(parent: str, schema: str | None, relname: str, alias: str | None, expected: bool) -> None:
    assert SQLCompleter._matches_parent(parent, schema, relname, alias) is expected


def test_copy_other_schemas_from_preserves_non_current_metadata() -> None:
    source = SQLCompleter()
    source.load_schema_metadata(
        schema='other',
        table_columns={'users': ['*', 'id', 'email']},
        foreign_keys={'tables': {}, 'relations': []},
        enum_values={},
        functions={'fn_foo': None},
        procedures={},
    )
    # Also populate the source's "current" schema; it should NOT be copied.
    source.load_schema_metadata(
        schema='current',
        table_columns={'stale_current': ['*']},
        foreign_keys={'tables': {}, 'relations': []},
        enum_values={},
        functions={},
        procedures={},
    )

    dest = SQLCompleter()
    dest.set_dbname('current')
    dest.extend_schemata('current')

    dest.copy_other_schemas_from(source, exclude='current')

    assert 'other' in dest.dbmetadata['tables']
    assert dest.dbmetadata['tables']['other'] == {'users': ['*', 'id', 'email']}
    assert dest.dbmetadata['functions']['other'] == {'fn_foo': None}
    # The excluded schema is not overwritten with stale source data.
    assert dest.dbmetadata['tables']['current'] == {}
    # Completion lookups pick up the copied names.
    assert 'users' in dest.all_completions
    assert 'email' in dest.all_completions
    assert 'fn_foo' in dest.all_completions


def test_copy_other_schemas_from_does_not_overwrite_existing_dest() -> None:
    source = SQLCompleter()
    source.load_schema_metadata(
        schema='shared',
        table_columns={'from_source': ['*']},
        foreign_keys={'tables': {}, 'relations': []},
        enum_values={},
        functions={},
        procedures={},
    )

    dest = SQLCompleter()
    dest.set_dbname('current')
    dest.dbmetadata['tables']['shared'] = {'from_dest': ['*']}

    dest.copy_other_schemas_from(source, exclude='current')

    # Destination's existing data wins over source when a conflict exists.
    assert dest.dbmetadata['tables']['shared'] == {'from_dest': ['*']}


def test_load_schema_metadata_ignores_empty_schema() -> None:
    completer = SQLCompleter()

    completer.load_schema_metadata(
        schema='',
        table_columns={'users': ['*', 'id']},
        foreign_keys={'tables': {'users': []}, 'relations': [('users', 'id')]},
        enum_values={'users': {'status': ['pending']}},
        functions={'fn_users': None},
        procedures={'proc_users': None},
    )

    assert completer.dbmetadata['tables'] == {}
    assert completer.dbmetadata['views'] == {}
    assert completer.dbmetadata['functions'] == {}
    assert completer.dbmetadata['procedures'] == {}
    assert completer.dbmetadata['enum_values'] == {}
    assert completer.dbmetadata['foreign_keys'] == {}
    assert 'users' not in completer.all_completions
    assert 'fn_users' not in completer.all_completions



# type: ignore

from mycli.packages.string_utils import sanitize_terminal_title


def test_sanitize_terminal_title_strips_ansi_sequences() -> None:
    title = '\x1b[31mmycli\x1b[0m session'

    assert sanitize_terminal_title(title) == 'mycli session'


def test_sanitize_terminal_title_replaces_newlines_with_spaces() -> None:
    title = 'schema\nquery\r\nprompt'

    assert sanitize_terminal_title(title) == 'schema query prompt'


def test_sanitize_terminal_title_removes_control_characters() -> None:
    title = 'my\x00cl\ti\x1f title\x7f'

    assert sanitize_terminal_title(title) == 'mycli title'


def test_sanitize_terminal_title_preserves_printable_text() -> None:
    title = 'db-01 / reporting'

    assert sanitize_terminal_title(title) == 'db-01 / reporting'



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, cast

import mycli.main_modes.list_dsn as list_dsn_mode


@dataclass
class DummyCliArgs:
    verbose: int = 0


class DummyConfig:
    def __init__(self, value: dict[str, str] | Exception) -> None:
        self.value = value

    def __getitem__(self, key: str) -> dict[str, str]:
        assert key == 'alias_dsn'
        if isinstance(self.value, Exception):
            raise self.value
        return self.value


class DummyMyCli:
    def __init__(self, config: Any) -> None:
        self.config = config
        self.verbosity = 0


def main_list_dsn(mycli: DummyMyCli) -> int:
    return list_dsn_mode.main_list_dsn(cast(Any, mycli))


def test_main_list_dsn_lists_aliases_without_values(monkeypatch) -> None:
    secho_calls: list[tuple[str, bool | None, str | None]] = []
    mycli = DummyMyCli(DummyConfig({'prod': 'mysql://u:p@h/db', 'staging': 'mysql://u2:p2@h2/db2'}))

    monkeypatch.setattr(
        list_dsn_mode.click,
        'secho',
        lambda message, err=None, fg=None: secho_calls.append((message, err, fg)),
    )

    result = main_list_dsn(mycli)

    assert result == 0
    assert secho_calls == [
        ('prod', None, None),
        ('staging', None, None),
    ]


def test_main_list_dsn_lists_aliases_with_values_in_verbose_mode(monkeypatch) -> None:
    secho_calls: list[tuple[str, bool | None, str | None]] = []
    mycli = DummyMyCli(DummyConfig({'prod': 'mysql://u:p@h/db'}))
    mycli.verbosity = 1

    monkeypatch.setattr(
        list_dsn_mode.click,
        'secho',
        lambda message, err=None, fg=None: secho_calls.append((message, err, fg)),
    )

    result = main_list_dsn(mycli)

    assert result == 0
    assert secho_calls == [('prod : mysql://u:p@h/db', None, None)]


def test_main_list_dsn_reports_invalid_alias_section(monkeypatch) -> None:
    secho_calls: list[tuple[str, bool | None, str | None]] = []
    mycli = DummyMyCli(DummyConfig(KeyError('alias_dsn')))

    monkeypatch.setattr(
        list_dsn_mode.click,
        'secho',
        lambda message, err=None, fg=None: secho_calls.append((message, err, fg)),
    )

    result = main_list_dsn(mycli)

    assert result == 1
    assert secho_calls == [
        (
            'Invalid DSNs found in the config file. Please check the "[alias_dsn]" section in myclirc.',
            True,
            'red',
        )
    ]


def test_main_list_dsn_reports_other_config_errors(monkeypatch) -> None:
    secho_calls: list[tuple[str, bool | None, str | None]] = []
    mycli = DummyMyCli(DummyConfig(RuntimeError('boom')))

    monkeypatch.setattr(
        list_dsn_mode.click,
        'secho',
        lambda message, err=None, fg=None: secho_calls.append((message, err, fg)),
    )

    result = main_list_dsn(mycli)

    assert result == 1
    assert secho_calls == [('boom', True, 'red')]



# type: ignore

import os.path
from unittest.mock import patch

from prompt_toolkit.completion import Completion
from prompt_toolkit.document import Document
import pytest

import mycli.packages.special.main as special
from test.utils import pygments_below

metadata = {
    "users": ["id", "email", "first_name", "last_name"],
    "orders": ["id", "ordered_date", "status"],
    "select": ["id", "insert", "ABC"],
    "réveillé": ["id", "insert", "ABC"],
    "time_zone": ["Time_zone_id"],
    "time_zone_leap_second": ["Time_zone_id"],
    "time_zone_name": ["Time_zone_id"],
    "time_zone_transition": ["Time_zone_id"],
    "time_zone_transition_type": ["Time_zone_id"],
}


@pytest.fixture
def completer():
    import mycli.sqlcompleter as sqlcompleter

    comp = sqlcompleter.SQLCompleter(smart_completion=True)

    tables, columns = [], []

    for table, cols in metadata.items():
        tables.append((table,))
        columns.extend([(table, col) for col in cols])

    databases = ["test", "test 2"]

    for db in databases:
        comp.extend_schemata(db)
    comp.extend_database_names(databases)
    comp.set_dbname("test")
    comp.extend_relations(tables, kind="tables")
    comp.extend_columns(columns, kind="tables")
    comp.extend_enum_values([("orders", "status", ["pending", "shipped"])])
    comp.extend_special_commands(special.COMMANDS)

    return comp


@pytest.fixture
def empty_completer():
    import mycli.sqlcompleter as sqlcompleter

    comp = sqlcompleter.SQLCompleter(smart_completion=True)

    tables, columns = [], []

    for table, cols in metadata.items():
        tables.append((table,))
        columns.extend([(table, col) for col in cols])

    db = 'empty'

    comp.extend_schemata(db)
    comp.extend_database_names([db])
    comp.set_dbname(db)
    comp.extend_special_commands(special.COMMANDS)

    return comp


@pytest.fixture
def complete_event():
    from unittest.mock import Mock

    return Mock()


def test_use_database_completion(completer, complete_event):
    text = "USE "
    position = len(text)
    special.register_special_command(
        ...,
        'use',
        '\\u [database]',
        'Change to a new database.',
        aliases=[special.SpecialCommandAlias('\\u', case_sensitive=False)],
    )
    result = completer.get_completions(Document(text=text, cursor_position=position), complete_event)
    assert list(result) == [
        Completion(text="test", start_position=0),
        Completion(text="`test 2`", start_position=0),
    ]


def test_special_name_completion(completer, complete_event):
    text = "\\d"
    position = len("\\d")
    result = completer.get_completions(Document(text=text, cursor_position=position), complete_event)
    assert list(result) == [Completion(text="\\dt", start_position=-2)]


def test_empty_string_completion(completer, complete_event):
    text = ""
    position = 0
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert list(map(Completion, completer.special_commands + completer.keywords)) == result


def test_select_keyword_completion(completer, complete_event):
    text = "SEL"
    position = len("SEL")
    result = completer.get_completions(Document(text=text, cursor_position=position), complete_event)
    assert list(result) == [
        Completion(text='SELECT', start_position=-3),
        Completion(text='SERIAL', start_position=-3),
        Completion(text='MASTER_LOG_FILE', start_position=-3),
        Completion(text='MASTER_LOG_POS', start_position=-3),
        Completion(text='MASTER_TLS_CIPHERSUITES', start_position=-3),
        Completion(text='MASTER_TLS_VERSION', start_position=-3),
        Completion(text='SCHEDULE', start_position=-3),
        Completion(text='SERIALIZABLE', start_position=-3),
    ]


def test_select_star(completer, complete_event):
    text = "SELECT * "
    position = len(text)
    result = completer.get_completions(Document(text=text, cursor_position=position), complete_event)
    assert list(result) == list(map(Completion, completer.keywords))


def test_introducer_completion(completer, complete_event):
    completer.extend_character_sets([('latin1',), ('utf8mb4',)])
    text = 'SELECT _'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    result_text = [item.text for item in result]
    assert '_latin1' in result_text
    assert '_utf8mb4' in result_text


def test_collation_completion(completer, complete_event):
    completer.extend_collations([('utf16le_bin',), ('utf8mb4_unicode_ci',)])
    text = 'SELECT "text" COLLATE '
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    result_text = [item.text for item in result]
    assert 'utf16le_bin' in result_text
    assert 'utf8mb4_unicode_ci' in result_text


def test_transcoding_completion_1(completer, complete_event):
    completer.extend_character_sets([('latin1',), ('utf8mb4',)])
    text = 'SELECT CONVERT("text" USING '
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    result_text = [item.text for item in result]
    assert 'latin1' in result_text
    assert 'utf8mb4' in result_text


def test_transcoding_completion_2(completer, complete_event):
    completer.extend_character_sets([('utf8mb3',), ('utf8mb4',)])
    text = 'SELECT CONVERT("text" USING u'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    result_text = [item.text for item in result]
    assert 'utf8mb3' in result_text
    assert 'utf8mb4' in result_text


def test_transcoding_completion_3(completer, complete_event):
    completer.extend_character_sets([('latin1',), ('utf8mb4',)])
    text = 'SELECT CAST("text" AS CHAR CHARACTER SET '
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    result_text = [item.text for item in result]
    assert 'latin1' in result_text
    assert 'utf8mb4' in result_text


def test_transcoding_completion_4(completer, complete_event):
    completer.extend_character_sets([('utf8mb3',), ('utf8mb4',)])
    text = 'SELECT CAST("text" AS CHAR CHARACTER SET u'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    result_text = [item.text for item in result]
    assert 'utf8mb3' in result_text
    assert 'utf8mb4' in result_text


def test_where_transcoding_completion_1(completer, complete_event):
    completer.extend_character_sets([('latin1',), ('utf8mb4',)])
    text = 'SELECT * FROM users WHERE CONVERT(email USING '
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    result_text = [item.text for item in result]
    assert 'latin1' in result_text
    assert 'utf8mb4' in result_text


def test_where_transcoding_completion_2(completer, complete_event):
    completer.extend_character_sets([('latin1',), ('utf8mb4',)])
    text = 'SELECT * FROM users WHERE CAST(email AS CHAR CHARACTER SET '
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    result_text = [item.text for item in result]
    assert 'latin1' in result_text
    assert 'utf8mb4' in result_text


def test_table_completion(completer, complete_event):
    text = "SELECT * FROM "
    position = len(text)
    result = completer.get_completions(Document(text=text, cursor_position=position), complete_event)
    assert list(result) == [
        Completion(text="users", start_position=0),
        Completion(text="orders", start_position=0),
        Completion(text="`select`", start_position=0),
        Completion(text="`réveillé`", start_position=0),
        Completion(text="time_zone", start_position=0),
        Completion(text="time_zone_leap_second", start_position=0),
        Completion(text="time_zone_name", start_position=0),
        Completion(text="time_zone_transition", start_position=0),
        Completion(text="time_zone_transition_type", start_position=0),
        Completion(text="test", start_position=0),
        Completion(text="`test 2`", start_position=0),
    ]


def test_select_filtered_table_completion(completer, complete_event):
    text = "SELECT ABC FROM "
    position = len(text)
    result = completer.get_completions(Document(text=text, cursor_position=position), complete_event)
    assert list(result) == [
        Completion(text="`select`", start_position=0),
        Completion(text="`réveillé`", start_position=0),
        Completion(text="users", start_position=0),
        Completion(text="orders", start_position=0),
        Completion(text="time_zone", start_position=0),
        Completion(text="time_zone_leap_second", start_position=0),
        Completion(text="time_zone_name", start_position=0),
        Completion(text="time_zone_transition", start_position=0),
        Completion(text="time_zone_transition_type", start_position=0),
        Completion(text="test", start_position=0),
        Completion(text="`test 2`", start_position=0),
    ]


def test_sub_select_filtered_table_completion(completer, complete_event):
    text = "SELECT * FROM (SELECT ordered_date FROM "
    position = len(text)
    result = completer.get_completions(Document(text=text, cursor_position=position), complete_event)
    assert list(result) == [
        Completion(text="orders", start_position=0),
        Completion(text="users", start_position=0),
        Completion(text="`select`", start_position=0),
        Completion(text="`réveillé`", start_position=0),
        Completion(text="time_zone", start_position=0),
        Completion(text="time_zone_leap_second", start_position=0),
        Completion(text="time_zone_name", start_position=0),
        Completion(text="time_zone_transition", start_position=0),
        Completion(text="time_zone_transition_type", start_position=0),
        Completion(text="test", start_position=0),
        Completion(text="`test 2`", start_position=0),
    ]


def test_enum_value_completion(completer, complete_event):
    text = "SELECT * FROM orders WHERE status = "
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="'pending'", start_position=0),
        Completion(text="'shipped'", start_position=0),
    ]


def test_function_name_completion(completer, complete_event):
    text = "SELECT MA"
    position = len("SELECT MA")
    result = completer.get_completions(Document(text=text, cursor_position=position), complete_event)
    assert list(result) == [
        Completion(text='MAX', start_position=-2),
        Completion(text='MATCH', start_position=-2),
        Completion(text='MAKEDATE', start_position=-2),
        Completion(text='MAKETIME', start_position=-2),
        Completion(text='MAKE_SET', start_position=-2),
        Completion(text='MASTER_POS_WAIT', start_position=-2),
        Completion(text='email', start_position=-2),
    ]


def test_suggested_column_names(completer, complete_event):
    """Suggest column and function names when selecting from table.

    :param completer:
    :param complete_event:
    :return:

    """
    text = "SELECT  from users"
    position = len("SELECT ")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == list(
        [
            Completion(text="*", start_position=0),
            Completion(text="id", start_position=0),
            Completion(text="email", start_position=0),
            Completion(text="first_name", start_position=0),
            Completion(text="last_name", start_position=0),
        ]
        + list(map(Completion, completer.functions))
        + [Completion(text="users", start_position=0)]
    )


def test_suggested_column_names_empty_db(empty_completer, complete_event):
    """Suggest * and function when selecting from no-table db.

    :param empty_completer:
    :param complete_event:
    :return:

    """
    text = "SELECT "
    position = len("SELECT ")
    result = list(empty_completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == list(
        [
            Completion(text="*", start_position=0),
        ]
        + list(map(Completion, empty_completer.functions))
    )


def test_suggested_column_names_in_function(completer, complete_event):
    """Suggest column and function names when selecting multiple columns from
    table.

    :param completer:
    :param complete_event:
    :return:

    """
    text = "SELECT MAX( from users"
    position = len("SELECT MAX(")
    result = completer.get_completions(Document(text=text, cursor_position=position), complete_event)
    assert list(result) == [
        Completion(text="*", start_position=0),
        Completion(text="id", start_position=0),
        Completion(text="email", start_position=0),
        Completion(text="first_name", start_position=0),
        Completion(text="last_name", start_position=0),
    ]


def test_suggested_column_names_with_table_dot(completer, complete_event):
    """Suggest column names on table name and dot.

    :param completer:
    :param complete_event:
    :return:

    """
    text = "SELECT users. from users"
    position = len("SELECT users.")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="*", start_position=0),
        Completion(text="id", start_position=0),
        Completion(text="email", start_position=0),
        Completion(text="first_name", start_position=0),
        Completion(text="last_name", start_position=0),
    ]


def test_suggested_column_names_with_alias(completer, complete_event):
    """Suggest column names on table alias and dot.

    :param completer:
    :param complete_event:
    :return:

    """
    text = "SELECT u. from users u"
    position = len("SELECT u.")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="*", start_position=0),
        Completion(text="id", start_position=0),
        Completion(text="email", start_position=0),
        Completion(text="first_name", start_position=0),
        Completion(text="last_name", start_position=0),
    ]


def test_suggested_multiple_column_names(completer, complete_event):
    """Suggest column and function names when selecting multiple columns from
    table.

    :param completer:
    :param complete_event:
    :return:

    """
    text = "SELECT id,  from users u"
    position = len("SELECT id, ")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == list(
        [
            Completion(text="*", start_position=0),
            Completion(text="id", start_position=0),
            Completion(text="email", start_position=0),
            Completion(text="first_name", start_position=0),
            Completion(text="last_name", start_position=0),
        ]
        + list(map(Completion, completer.functions))
        + [Completion(text="u", start_position=0)]
    )


def test_suggested_multiple_column_names_with_alias(completer, complete_event):
    """Suggest column names on table alias and dot when selecting multiple
    columns from table.

    :param completer:
    :param complete_event:
    :return:

    """
    text = "SELECT u.id, u. from users u"
    position = len("SELECT u.id, u.")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="*", start_position=0),
        Completion(text="id", start_position=0),
        Completion(text="email", start_position=0),
        Completion(text="first_name", start_position=0),
        Completion(text="last_name", start_position=0),
    ]


def test_suggested_multiple_column_names_with_dot(completer, complete_event):
    """Suggest column names on table names and dot when selecting multiple
    columns from table.

    :param completer:
    :param complete_event:
    :return:

    """
    text = "SELECT users.id, users. from users u"
    position = len("SELECT users.id, users.")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="*", start_position=0),
        Completion(text="id", start_position=0),
        Completion(text="email", start_position=0),
        Completion(text="first_name", start_position=0),
        Completion(text="last_name", start_position=0),
    ]


def test_suggested_aliases_after_on(completer, complete_event):
    text = "SELECT u.name, o.id FROM users u JOIN orders o ON "
    position = len("SELECT u.name, o.id FROM users u JOIN orders o ON ")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="u", start_position=0),
        Completion(text="o", start_position=0),
    ]


def test_suggested_aliases_after_on_right_side(completer, complete_event):
    text = "SELECT u.name, o.id FROM users u JOIN orders o ON o.user_id = "
    position = len("SELECT u.name, o.id FROM users u JOIN orders o ON o.user_id = ")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="u", start_position=0),
        Completion(text="o", start_position=0),
    ]


def test_suggested_tables_after_on(completer, complete_event):
    text = "SELECT users.name, orders.id FROM users JOIN orders ON "
    position = len("SELECT users.name, orders.id FROM users JOIN orders ON ")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="users", start_position=0),
        Completion(text="orders", start_position=0),
    ]


def test_suggested_tables_after_on_right_side(completer, complete_event):
    text = "SELECT users.name, orders.id FROM users JOIN orders ON orders.user_id = "
    position = len("SELECT users.name, orders.id FROM users JOIN orders ON orders.user_id = ")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="users", start_position=0),
        Completion(text="orders", start_position=0),
    ]


def test_table_names_after_from(completer, complete_event):
    text = "SELECT * FROM "
    position = len("SELECT * FROM ")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="users", start_position=0),
        Completion(text="orders", start_position=0),
        Completion(text="`select`", start_position=0),
        Completion(text="`réveillé`", start_position=0),
        Completion(text="time_zone", start_position=0),
        Completion(text="time_zone_leap_second", start_position=0),
        Completion(text="time_zone_name", start_position=0),
        Completion(text="time_zone_transition", start_position=0),
        Completion(text="time_zone_transition_type", start_position=0),
        Completion(text="test", start_position=0),
        Completion(text="`test 2`", start_position=0),
    ]


def test_table_names_leading_partial(completer, complete_event):
    text = "SELECT * FROM time_zone"
    position = len("SELECT * FROM time_zone")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="time_zone", start_position=-9),
        Completion(text="time_zone_name", start_position=-9),
        Completion(text="time_zone_transition", start_position=-9),
        Completion(text="time_zone_leap_second", start_position=-9),
        Completion(text="time_zone_transition_type", start_position=-9),
    ]


def test_table_names_inter_partial(completer, complete_event):
    text = "SELECT * FROM time_leap"
    position = len("SELECT * FROM time_leap")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="time_zone_leap_second", start_position=-9),
        Completion(text='time_zone_name', start_position=-9),
        Completion(text='time_zone_transition', start_position=-9),
        Completion(text='time_zone_transition_type', start_position=-9),
    ]


def test_table_names_fuzzy(completer, complete_event):
    text = "SELECT * FROM tim_leap"
    position = len("SELECT * FROM tim_leap")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="time_zone_leap_second", start_position=-8),
    ]


def test_auto_escaped_col_names(completer, complete_event):
    text = "SELECT  from `select`"
    position = len("SELECT ")
    result = [x.text for x in completer.get_completions(Document(text=text, cursor_position=position), complete_event)]
    expected = (
        [
            "*",
            "id",
            "`insert`",
            "ABC",
        ]
        + completer.functions
        + ["select"]
    )
    assert result == expected


def test_un_escaped_table_names(completer, complete_event):
    text = "SELECT  from réveillé"
    position = len("SELECT ")
    result = [x.text for x in completer.get_completions(Document(text=text, cursor_position=position), complete_event)]
    assert result == [
        "*",
        "id",
        "`insert`",
        "ABC",
    ] + completer.functions + ["réveillé"]


# todo: the fixtures are insufficient; the database name should also appear in the result
def test_grant_on_suggets_tables_and_schemata(completer, complete_event):
    text = "GRANT ALL ON "
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="test", start_position=0),
        Completion(text="`test 2`", start_position=0),
        Completion(text='users', start_position=0),
        Completion(text='orders', start_position=0),
        Completion(text='`select`', start_position=0),
        Completion(text='`réveillé`', start_position=0),
        Completion(text='time_zone', start_position=0),
        Completion(text='time_zone_leap_second', start_position=0),
        Completion(text='time_zone_name', start_position=0),
        Completion(text='time_zone_transition', start_position=0),
        Completion(text='time_zone_transition_type', start_position=0),
    ]


# todo: this test belongs more logically in test_naive_completion.py, but it didn't work there:
#       multiple completion candidates were not suggested.
def test_deleted_keyword_completion(completer, complete_event):
    text = "exi"
    position = len("exi")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text="exit", start_position=-3),
        Completion(text='exists', start_position=-3),
        Completion(text='explain', start_position=-3),
        Completion(text='expire', start_position=-3),
    ]


def test_numbers_no_completion(completer, complete_event):
    text = "SELECT COUNT(1) FROM time_zone WHERE Time_zone_id = 1"
    position = len("SELECT COUNT(1) FROM time_zone WHERE Time_zone_id = 1")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == []  # ie not INT1


def dummy_list_path(dir_name):
    dirs = {
        "/": [
            "dir1",
            "file1.sql",
            "file2.sql",
        ],
        "/dir1": [
            "subdir1",
            "subfile1.sql",
            "subfile2.sql",
        ],
        "/dir1/subdir1": [
            "lastfile.sql",
        ],
    }
    return dirs.get(dir_name, [])


@patch("mycli.packages.filepaths.list_path", new=dummy_list_path)
@pytest.mark.parametrize(
    "text,expected",
    [
        ('source ', [('/', 0), ('~', 0), ('.', 0), ('..', 0)]),
        ("source /", [("dir1", 0), ("file1.sql", 0), ("file2.sql", 0)]),
        ("source /dir1/", [("subdir1", 0), ("subfile1.sql", 0), ("subfile2.sql", 0)]),
        ("source /dir1/subdir1/", [("lastfile.sql", 0)]),
    ],
)
@pytest.mark.skipif(os.name == 'nt', reason='todo: unknown')
def test_file_name_completion(completer, complete_event, text, expected):
    position = len(text)
    special.register_special_command(
        ...,
        'source',
        '\\. <filename>',
        'Execute commands from file.',
        aliases=[special.SpecialCommandAlias('\\.', case_sensitive=False)],
    )
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    expected = [Completion(txt, pos) for txt, pos in expected]
    assert result == expected


def test_auto_case_heuristic(completer, complete_event):
    text = "select json_v"
    position = len("select json_v")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert [x.text for x in result] == [
        'json_value',
        'json_valid',
    ]


def test_create_table_like_completion(completer, complete_event):
    text = "CREATE TABLE foo LIKE ti"
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert [x.text for x in result] == [
        'time_zone',
        'time_zone_name',
        'time_zone_transition',
        'time_zone_leap_second',
        'time_zone_transition_type',
    ]


def test_source_eager_completion(completer, complete_event, tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    os.mkdir('doc')
    text = "source do"
    position = len(text)
    script_filename = 'do_these_statements.sql'
    f = open(script_filename, 'w')
    f.close()
    special.register_special_command(
        ...,
        'source',
        '\\. <filename>',
        'Execute commands from file.',
        aliases=[special.SpecialCommandAlias('\\.', case_sensitive=False)],
    )
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    success = True
    error = 'unknown'
    try:
        assert [x.text for x in result] == [
            script_filename,
            'doc/',
        ]
    except AssertionError as e:
        success = False
        error = e
    if os.path.exists(script_filename):
        os.remove(script_filename)
    if not success:
        raise AssertionError(error)


def test_source_leading_dot_suggestions_completion(completer, complete_event, tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    os.mkdir('doc')
    text = "source ./do"
    position = len(text)
    script_filename = 'do_these_statements.sql'
    f = open(script_filename, 'w')
    f.close()
    special.register_special_command(
        ...,
        'source',
        '\\. <filename>',
        'Execute commands from file.',
        aliases=[special.SpecialCommandAlias('\\.', case_sensitive=False)],
    )
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    success = True
    error = 'unknown'
    try:
        assert [x.text for x in result] == [
            script_filename,
            'doc/',
        ]
    except AssertionError as e:
        success = False
        error = e
    if os.path.exists(script_filename):
        os.remove(script_filename)
    if not success:
        raise AssertionError(error)


def test_string_no_completion(completer, complete_event):
    text = 'select "json'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == []


def test_string_no_completion_single_quote(completer, complete_event):
    text = "select 'json"
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == []


def test_string_no_completion_spaces(completer, complete_event):
    text = 'select "nocomplete json'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == []


def test_string_no_completion_spaces_inner_1(completer, complete_event):
    text = 'select "json nocomplete'
    position = len('select "json')
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == []


def test_string_no_completion_spaces_inner_2(completer, complete_event):
    text = 'select "json nocomplete'
    position = len('select "json ')
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == []


def test_backticked_column_completion(completer, complete_event):
    text = 'select `Tim'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        # todo it would be nicer if the column names sorted to the top
        Completion(text='`time`', start_position=-4),
        Completion(text='`timediff`', start_position=-4),
        Completion(text='`timestamp`', start_position=-4),
        Completion(text='`time_format`', start_position=-4),
        Completion(text='`time_to_sec`', start_position=-4),
        Completion(text='`Time_zone_id`', start_position=-4),
        Completion(text='`timestampadd`', start_position=-4),
        Completion(text='`timestampdiff`', start_position=-4),
        Completion(text='`datetime`', start_position=-4),
        Completion(text='`optimize`', start_position=-4),
        Completion(text='`optimizer_costs`', start_position=-4),
        Completion(text='`utc_time`', start_position=-4),
        Completion(text='`utc_timestamp`', start_position=-4),
        Completion(text='`current_time`', start_position=-4),
        Completion(text='`current_timestamp`', start_position=-4),
        Completion(text='`localtime`', start_position=-4),
        Completion(text='`localtimestamp`', start_position=-4),
        Completion(text='`password_lock_time`', start_position=-4),
    ]


def test_backticked_column_completion_component(completer, complete_event):
    text = 'select `com'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        # todo it would be nicer if "comment" sorted to the top because it is a column name,
        #      and because it is a reserved word
        Completion(text='`commit`', start_position=-4),
        Completion(text='`comment`', start_position=-4),
        Completion(text='`compact`', start_position=-4),
        Completion(text='`compress`', start_position=-4),
        Completion(text='`committed`', start_position=-4),
        Completion(text='`component`', start_position=-4),
        Completion(text='`completion`', start_position=-4),
        Completion(text='`compressed`', start_position=-4),
        Completion(text='`compression`', start_position=-4),
        Completion(text='`column`', start_position=-4),
        Completion(text='`column_format`', start_position=-4),
        Completion(text='`column_name`', start_position=-4),
        Completion(text='`columns`', start_position=-4),
        Completion(text='`second_microsecond`', start_position=-4),
        Completion(text='`uncommitted`', start_position=-4),
    ]


def test_backticked_column_completion_two_character(completer, complete_event):
    text = 'select `f'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    expected = [
        # todo it would be nicer if the column name "first_name" sorted to the top
        Completion(text='`for`', start_position=-2),
        Completion(text='`from`', start_position=-2),
        Completion(text='`fast`', start_position=-2),
        Completion(text='`file`', start_position=-2),
        Completion(text='`full`', start_position=-2),
        Completion(text='`floor`', start_position=-2),
        Completion(text='`false`', start_position=-2),
        Completion(text='`field`', start_position=-2),
        Completion(text='`fixed`', start_position=-2),
        Completion(text='`float`', start_position=-2),
        Completion(text='`fetch`', start_position=-2),
        Completion(text='`files`', start_position=-2),
        Completion(text='`first`', start_position=-2),
        Completion(text='`flush`', start_position=-2),
        Completion(text='`force`', start_position=-2),
        Completion(text='`found`', start_position=-2),
        Completion(text='`format`', start_position=-2),
        Completion(text='`float4`', start_position=-2),
        Completion(text='`float8`', start_position=-2),
        Completion(text='`factor`', start_position=-2),
        Completion(text='`faults`', start_position=-2),
        Completion(text='`fields`', start_position=-2),
        Completion(text='`filter`', start_position=-2),
        Completion(text='`finish`', start_position=-2),
        Completion(text='`follows`', start_position=-2),
        Completion(text='`foreign`', start_position=-2),
        Completion(text='`fulltext`', start_position=-2),
        Completion(text='`function`', start_position=-2),
        Completion(text='`from_days`', start_position=-2),
        Completion(text='`file_name`', start_position=-2),
        Completion(text='`following`', start_position=-2),
        Completion(text='`first_name`', start_position=-2),
        Completion(text='`found_rows`', start_position=-2),
        Completion(text='`find_in_set`', start_position=-2),
        Completion(text='`first_value`', start_position=-2),
        Completion(text='`from_base64`', start_position=-2),
        Completion(text='`from_vector`', start_position=-2),
        Completion(text='`file_format`', start_position=-2),
        Completion(text='`file_prefix`', start_position=-2),
        Completion(text='`foreign key`', start_position=-2),
        Completion(text='`format_bytes`', start_position=-2),
        Completion(text='`file_pattern`', start_position=-2),
        Completion(text='`from_unixtime`', start_position=-2),
        Completion(text='`file_block_size`', start_position=-2),
        Completion(text='`format_pico_time`', start_position=-2),
        Completion(text='`failed_login_attempts`', start_position=-2),
        Completion(text='`left join`', start_position=-2),
        Completion(text='`after`', start_position=-2),
        Completion(text='`before`', start_position=-2),
        Completion(text='`default`', start_position=-2),
        Completion(text='`default_auth`', start_position=-2),
        Completion(text='`definer`', start_position=-2),
        Completion(text='`definition`', start_position=-2),
        Completion(text='`enforced`', start_position=-2),
        Completion(text='`if`', start_position=-2),
        Completion(text='`infile`', start_position=-2),
        Completion(text='`left`', start_position=-2),
        Completion(text='`logfile`', start_position=-2),
        Completion(text='`of`', start_position=-2),
        Completion(text='`off`', start_position=-2),
        Completion(text='`offset`', start_position=-2),
        Completion(text='`outfile`', start_position=-2),
        Completion(text='`profile`', start_position=-2),
        Completion(text='`profiles`', start_position=-2),
        Completion(text='`reference`', start_position=-2),
        Completion(text='`references`', start_position=-2),
    ]

    if pygments_below("2.20"):
        for newer in [
            Completion(text='`file_format`', start_position=-2),
            Completion(text='`file_name`', start_position=-2),
            Completion(text='`file_pattern`', start_position=-2),
            Completion(text='`file_prefix`', start_position=-2),
            Completion(text='`files`', start_position=-2),
            Completion(text='`from_vector`', start_position=-2),
        ]:
            expected.remove(newer)

    assert result == expected


def test_backticked_column_completion_three_character(completer, complete_event):
    text = 'select `fi'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    expected = [
        # todo it would be nicer if the column name "first_name" sorted to the top
        Completion(text='`file`', start_position=-3),
        Completion(text='`field`', start_position=-3),
        Completion(text='`fixed`', start_position=-3),
        Completion(text='`files`', start_position=-3),
        Completion(text='`first`', start_position=-3),
        Completion(text='`fields`', start_position=-3),
        Completion(text='`filter`', start_position=-3),
        Completion(text='`finish`', start_position=-3),
        Completion(text='`file_name`', start_position=-3),
        Completion(text='`first_name`', start_position=-3),
        Completion(text='`find_in_set`', start_position=-3),
        Completion(text='`first_value`', start_position=-3),
        Completion(text='`file_format`', start_position=-3),
        Completion(text='`file_prefix`', start_position=-3),
        Completion(text='`file_pattern`', start_position=-3),
        Completion(text='`file_block_size`', start_position=-3),
        Completion(text='`definer`', start_position=-3),
        Completion(text='`definition`', start_position=-3),
        Completion(text='`failed_login_attempts`', start_position=-3),
        Completion(text='`foreign`', start_position=-3),
        Completion(text='`infile`', start_position=-3),
        Completion(text='`logfile`', start_position=-3),
        Completion(text='`outfile`', start_position=-3),
        Completion(text='`profile`', start_position=-3),
        Completion(text='`profiles`', start_position=-3),
        Completion(text='`foreign key`', start_position=-3),
    ]

    if pygments_below("2.20"):
        for newer in [
            Completion(text='`files`', start_position=-3),
            Completion(text='`file_name`', start_position=-3),
            Completion(text='`file_format`', start_position=-3),
            Completion(text='`file_pattern`', start_position=-3),
            Completion(text='`file_prefix`', start_position=-3),
        ]:
            expected.remove(newer)

    assert result == expected


def test_backticked_column_completion_four_character(completer, complete_event):
    text = 'select `fir'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        # todo it would be nicer if the column name "first_name" sorted to the top
        Completion(text='`first`', start_position=-4),
        Completion(text='`first_name`', start_position=-4),
        Completion(text='`first_value`', start_position=-4),
        Completion(text='`definer`', start_position=-4),
        Completion(text='`filter`', start_position=-4),
    ]


def test_backticked_table_completion_required(completer, complete_event):
    text = 'select ABC from `rév'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text='`réveillé`', start_position=-4),
    ]


def test_backticked_table_completion_not_required(completer, complete_event):
    text = 'select * from `t'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [
        Completion(text='`test`', start_position=-2),
        Completion(text='`test 2`', start_position=-2),
        Completion(text='`time_zone`', start_position=-2),
        Completion(text='`time_zone_name`', start_position=-2),
        Completion(text='`time_zone_transition`', start_position=-2),
        Completion(text='`time_zone_leap_second`', start_position=-2),
        Completion(text='`time_zone_transition_type`', start_position=-2),
    ]


def test_string_no_completion_backtick(completer, complete_event):
    text = 'select * from "`t'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == []


# todo this shouldn't suggest anything but the space resets the logic
# and it completes on "bar" alone
@pytest.mark.xfail
def test_backticked_no_completion_spaces(completer, complete_event):
    text = 'select * from `nocomplete bar'
    position = len(text)
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == []


# Foreign key completion tests
@pytest.fixture
def fk_completer():
    """SQLCompleter with tables and a FK relationship.

    Schema:
        orders (id, user_id, ordered_date, status)  FK: user_id -> users.id
        users  (id, email, first_name)
        tags   (id, name)                           no FK
    """
    import mycli.packages.special.main as special
    import mycli.sqlcompleter as sqlcompleter

    comp = sqlcompleter.SQLCompleter(smart_completion=True)

    tables = [("orders",), ("users",), ("tags",)]
    columns = [
        ("orders", "id"),
        ("orders", "user_id"),
        ("orders", "ordered_date"),
        ("orders", "status"),
        ("users", "id"),
        ("users", "email"),
        ("users", "first_name"),
        ("tags", "id"),
        ("tags", "name"),
    ]
    fk_data = [("orders", "user_id", "users", "id")]

    comp.extend_schemata("test")
    comp.extend_database_names(["test"])
    comp.set_dbname("test")
    comp.extend_relations(tables, kind="tables")
    comp.extend_columns(columns, kind="tables")
    comp.extend_foreign_keys(fk_data)
    comp.extend_special_commands(special.COMMANDS)

    return comp


def test_extend_foreign_keys_stores_relation(fk_completer):
    relations = fk_completer.dbmetadata["foreign_keys"]["test"]["relations"]
    assert ("orders", "user_id", "users", "id") in relations


def test_extend_foreign_keys_stores_bidirectional_table_map(fk_completer):
    tables_map = fk_completer.dbmetadata["foreign_keys"]["test"]["tables"]
    assert "users" in tables_map["orders"]
    assert "orders" in tables_map["users"]


def test_extend_foreign_keys_unrelated_table_absent_from_map(fk_completer):
    tables_map = fk_completer.dbmetadata["foreign_keys"]["test"]["tables"]
    assert "tags" not in tables_map


def test_fk_join_conditions_with_aliases(fk_completer):
    conditions = fk_completer._fk_join_conditions([(None, "orders", "o"), (None, "users", "u")])
    assert conditions == ["o.user_id = u.id"]


def test_fk_join_conditions_without_aliases(fk_completer):
    conditions = fk_completer._fk_join_conditions([(None, "orders", None), (None, "users", None)])
    assert conditions == ["orders.user_id = users.id"]


def test_fk_join_conditions_single_table_yields_nothing(fk_completer):
    conditions = fk_completer._fk_join_conditions([(None, "orders", "o")])
    assert conditions == []


def test_fk_join_conditions_unrelated_tables_yields_nothing(fk_completer):
    conditions = fk_completer._fk_join_conditions([(None, "orders", "o"), (None, "tags", "t")])
    assert conditions == []


def test_join_suggests_fk_table_before_unrelated(fk_completer, complete_event):
    text = "SELECT * FROM orders JOIN "
    result = [c.text for c in fk_completer.get_completions(Document(text=text, cursor_position=len(text)), complete_event)]
    assert "users" in result
    assert "tags" in result
    assert result.index("users") < result.index("tags")


def test_join_fk_lookup_is_bidirectional(fk_completer, complete_event):
    text = "SELECT * FROM users JOIN "
    result = [c.text for c in fk_completer.get_completions(Document(text=text, cursor_position=len(text)), complete_event)]
    assert "orders" in result
    assert "tags" in result
    assert result.index("orders") < result.index("tags")


def test_join_unrelated_table_still_suggests_all_tables(fk_completer, complete_event):
    text = "SELECT * FROM tags JOIN "
    result = [c.text for c in fk_completer.get_completions(Document(text=text, cursor_position=len(text)), complete_event)]
    assert "orders" in result
    assert "users" in result


def test_on_suggests_fk_condition_with_aliases(fk_completer, complete_event):
    text = "SELECT * FROM orders o JOIN users u ON "
    result = [c.text for c in fk_completer.get_completions(Document(text=text, cursor_position=len(text)), complete_event)]
    assert "o.user_id = u.id" in result


def test_on_suggests_fk_condition_without_aliases(fk_completer, complete_event):
    text = "SELECT * FROM orders JOIN users ON "
    result = [c.text for c in fk_completer.get_completions(Document(text=text, cursor_position=len(text)), complete_event)]
    assert "orders.user_id = users.id" in result


def test_on_fk_condition_appears_before_aliases(fk_completer, complete_event):
    text = "SELECT * FROM orders o JOIN users u ON "
    result = [c.text for c in fk_completer.get_completions(Document(text=text, cursor_position=len(text)), complete_event)]
    assert result.index("o.user_id = u.id") < result.index("o")


def test_on_no_fk_condition_for_unrelated_join(fk_completer, complete_event):
    text = "SELECT * FROM orders o JOIN tags t ON "
    result = [c.text for c in fk_completer.get_completions(Document(text=text, cursor_position=len(text)), complete_event)]
    assert not any("=" in r for r in result)
    assert "o" in result
    assert "t" in result


def test_on_partial_text_filters_fk_condition(fk_completer, complete_event):
    text = "SELECT * FROM orders JOIN users ON ord"
    result = [c.text for c in fk_completer.get_completions(Document(text=text, cursor_position=len(text)), complete_event)]
    assert "orders.user_id = users.id" in result


def test_fk_reserved_column_names_are_escaped():
    """FK columns that are reserved words or need quoting must be backtick-escaped."""
    import mycli.sqlcompleter as sqlcompleter

    comp = sqlcompleter.SQLCompleter(smart_completion=True)
    comp.extend_schemata("test")
    comp.set_dbname("test")
    comp.extend_foreign_keys([("orders", "order", "users", "select")])

    relations = comp.dbmetadata["foreign_keys"]["test"]["relations"]
    assert ("orders", "`order`", "users", "`select`") in relations

    conditions = comp._fk_join_conditions([(None, "orders", "o"), (None, "users", "u")])
    assert conditions == ["o.`order` = u.`select`"]


def test_fk_conditions_ignore_cross_schema_tables(fk_completer):
    """Tables qualified with a foreign schema are excluded from FK condition generation."""
    tables = [("other_db", "orders", "o"), (None, "users", "u")]
    conditions = fk_completer._fk_join_conditions(tables)
    assert conditions == []


def test_join_priority_ignores_cross_schema_table(fk_completer, complete_event):
    """Schema-qualified tables in FROM do not trigger FK priority using current-db metadata."""
    text = "SELECT * FROM other_db.orders JOIN "
    result_cross_schema = [c.text for c in fk_completer.get_completions(Document(text=text, cursor_position=len(text)), complete_event)]
    # A table with no FK relationships at all should give the same ordering,
    # confirming that no FK priority was applied for the cross-schema table.
    text_no_fk = "SELECT * FROM tags JOIN "
    result_no_fk = [
        c.text for c in fk_completer.get_completions(Document(text=text_no_fk, cursor_position=len(text_no_fk)), complete_event)
    ]
    assert result_cross_schema == result_no_fk



# type: ignore

"""Steps for behavioral style tests are defined in this module.

Each step is defined by the string decorating it. This string is used
to call the step in "*.feature" file.

"""

import datetime
import tempfile
from textwrap import dedent

from behave import then, when
import wrappers

from test.utils import TEMPFILE_PREFIX


@when("we run dbcli")
def step_run_cli(context):
    wrappers.run_cli(context)


@when("we wait for prompt")
def step_wait_prompt(context):
    wrappers.wait_prompt(context)


@when('we send "ctrl + d"')
def step_ctrl_d(context):
    """Send Ctrl + D to hopefully exit."""
    context.cli.sendcontrol("d")
    context.exit_sent = True


@when('we send "ctrl + o, ctrl + d"')
def step_ctrl_o_ctrl_d(context):
    """Send ctrl + o, ctrl + d to insert the quoted date."""
    context.cli.send("SELECT ")
    context.cli.sendcontrol("o")
    context.cli.sendcontrol("d")
    context.cli.send(" AS dt")
    context.cli.sendline("")


@when(r'we send "\?" command')
def step_send_help(context):
    r"""Send \?

    to see help.

    """
    context.cli.sendline("\\?")
    wrappers.expect_exact(context, context.conf["pager_boundary"] + "\r\n", timeout=5)


@when("we send source command")
def step_send_source_command(context):
    with tempfile.NamedTemporaryFile(prefix=TEMPFILE_PREFIX) as f:
        f.write(b"\\?")
        f.flush()
        context.cli.sendline(f"\\. {f.name}")
        wrappers.expect_exact(context, context.conf["pager_boundary"] + "\r\n", timeout=5)


@when("we run query to check application_name")
def step_check_application_name(context):
    context.cli.sendline(
        "SELECT 'found' FROM performance_schema.session_connect_attrs WHERE attr_name = 'program_name' AND attr_value = 'mycli' LIMIT 1"
    )


@then("we see found")
def step_see_found(context):
    expected = (
        dedent(
            """
            +-------+\r
            | found |\r
            +-------+\r
            | found |\r
            +-------+
            """
        ).strip()
        + '\r\n\r\n'
    )

    wrappers.expect_exact(
        context,
        context.conf["pager_boundary"] + '\r\n' + expected + context.conf["pager_boundary"],
        timeout=5,
    )


@then("we see the date")
def step_see_date(context):
    # There are some edge cases in which this test could fail,
    # such as running near midnight when the test database has
    # a different TZ setting than the system.
    date_str = datetime.datetime.now().strftime("%Y-%m-%d")
    expected = (
        dedent(
            f"""
            +------------+\r
            | dt         |\r
            +------------+\r
            | {date_str} |\r
            +------------+
            """
        ).strip()
        + '\r\n\r\n'
    )

    wrappers.expect_exact(
        context,
        context.conf["pager_boundary"] + '\r\n' + expected + context.conf["pager_boundary"],
        timeout=5,
    )


@then("we confirm the destructive warning")
def step_confirm_destructive_command(context):  # noqa
    """Confirm destructive command."""
    wrappers.expect_exact(context, "You're about to run a destructive command.\r\nDo you want to proceed? (y/n):", timeout=2)
    context.cli.sendline("y")


@when('we answer the destructive warning with "{confirmation}"')
def step_confirm_destructive_command(context, confirmation):  # noqa
    """Confirm destructive command."""
    wrappers.expect_exact(context, "You're about to run a destructive command.\r\nDo you want to proceed? (y/n):", timeout=2)
    context.cli.sendline(confirmation)


@then('we answer the destructive warning with invalid "{confirmation}" and see text "{text}"')
def step_confirm_destructive_command(context, confirmation, text):  # noqa
    """Confirm destructive command."""
    wrappers.expect_exact(context, "You're about to run a destructive command.\r\nDo you want to proceed? (y/n):", timeout=2)
    context.cli.sendline(confirmation)
    wrappers.expect_exact(context, text, timeout=2)
    # we must exit the Click loop, or the feature will hang
    context.cli.sendline("n")



from typing import IO, Generator

import sqlglot
import sqlparse

MAX_MULTILINE_BATCH_STATEMENT = 5000


def statements_from_filehandle(file_h: IO) -> Generator[tuple[str, int], None, None]:
    statements = ''
    line_counter = 0
    batch_counter = 0
    for batch_text in file_h:
        line_counter += 1
        if line_counter > MAX_MULTILINE_BATCH_STATEMENT:
            raise ValueError(f'Saw single input statement greater than {MAX_MULTILINE_BATCH_STATEMENT} lines; assuming a parsing error.')
        statements += batch_text
        try:
            tokens = sqlglot.tokenize(statements, read='mysql')
            if not tokens:
                continue
            # we don't yet handle changing the delimiter within the batch input
            if tokens[-1].text == ';':
                # The advantage of sqlparse for splitting is that it preserves the input.
                # https://github.com/tobymao/sqlglot/issues/2587#issuecomment-1823109501
                for statement in sqlparse.split(statements):
                    yield (statement, batch_counter)
                    batch_counter += 1
                statements = ''
                line_counter = 0
        except sqlglot.errors.TokenError:
            continue
    if statements:
        for statement in sqlparse.split(statements):
            yield (statement, batch_counter)
            batch_counter += 1



from types import SimpleNamespace
from typing import Any, cast

import pytest

from mycli.packages.ptoolkit import fzf as fzf_module
from mycli.packages.ptoolkit.history import FileHistoryWithTimestamp


class DummyHistory(FileHistoryWithTimestamp):
    def __init__(self, items: list[tuple[str, str]]) -> None:
        self._items = items

    def load_history_with_timestamp(self) -> list[tuple[str, str]]:
        return self._items


def make_event(history: Any) -> SimpleNamespace:
    buffer = SimpleNamespace(history=history, text='original', cursor_position=0)
    return SimpleNamespace(
        current_buffer=buffer,
        app=SimpleNamespace(),
    )


def test_fzf_init_and_is_available(monkeypatch) -> None:
    init_calls: list[bool] = []

    monkeypatch.setattr(fzf_module, 'which', lambda executable: '/usr/bin/fzf' if executable == 'fzf' else None)
    monkeypatch.setattr(fzf_module.FzfPrompt, '__init__', lambda self: init_calls.append(True))

    fzf = fzf_module.Fzf()

    assert fzf.executable == '/usr/bin/fzf'
    assert fzf.is_available() is True
    assert init_calls == [True]


def test_fzf_init_without_executable_skips_super(monkeypatch) -> None:
    init_calls: list[bool] = []

    monkeypatch.setattr(fzf_module, 'which', lambda executable: None)
    monkeypatch.setattr(fzf_module.FzfPrompt, '__init__', lambda self: init_calls.append(True))

    fzf = fzf_module.Fzf()

    assert fzf.executable is None
    assert fzf.is_available() is False
    assert init_calls == []


def test_search_history_falls_back_to_prompt_toolkit_search(monkeypatch) -> None:
    calls: list[dict[str, Any]] = []
    event = make_event(history=object())

    monkeypatch.setattr(
        fzf_module.search,
        'start_search',
        lambda **kwargs: calls.append(kwargs),
    )

    fzf_module.search_history(cast(Any, event), incremental=True)

    assert calls == [{'direction': fzf_module.search.SearchDirection.BACKWARD}]


def test_search_history_falls_back_when_fzf_unavailable_or_history_type_is_wrong(monkeypatch) -> None:
    calls: list[dict[str, Any]] = []
    unavailable_event = make_event(history=DummyHistory([]))
    wrong_history_event = make_event(history=[])

    class UnavailableFzf:
        def is_available(self) -> bool:
            return False

    monkeypatch.setattr(
        fzf_module.search,
        'start_search',
        lambda **kwargs: calls.append(kwargs),
    )

    monkeypatch.setattr(fzf_module, 'Fzf', UnavailableFzf)
    fzf_module.search_history(cast(Any, unavailable_event))

    class AvailableFzf:
        def is_available(self) -> bool:
            return True

    monkeypatch.setattr(fzf_module, 'Fzf', AvailableFzf)
    fzf_module.search_history(cast(Any, wrong_history_event))

    assert calls == [
        {'direction': fzf_module.search.SearchDirection.BACKWARD},
        {'direction': fzf_module.search.SearchDirection.BACKWARD},
    ]


def test_search_history_formats_preview_updates_buffer_and_deduplicates(monkeypatch) -> None:
    prompt_calls: list[dict[str, Any]] = []
    invalidated_apps: list[Any] = []

    history = DummyHistory([
        ('SELECT  1\nFROM dual', '2026-01-02 03:04:05.678'),
        ('SELECT 1 FROM dual', '2026-01-01 00:00:00'),
        ('SELECT 2', '2026-01-03 12:00:00'),
    ])
    event = make_event(history=history)

    class PromptingFzf:
        def is_available(self) -> bool:
            return True

        def prompt(self, items: list[str], fzf_options: str) -> list[str]:
            prompt_calls.append({'items': items, 'options': fzf_options})
            return [items[0]]

    monkeypatch.setattr(fzf_module, 'Fzf', PromptingFzf)
    monkeypatch.setattr(
        fzf_module,
        'which',
        lambda executable: '/usr/bin/pygmentize' if executable == 'pygmentize' else None,
    )
    monkeypatch.setattr(fzf_module, 'safe_invalidate_display', lambda app: invalidated_apps.append(app))

    fzf_module.search_history(
        cast(Any, event),
        highlight_preview=True,
        highlight_style='monokai style',
    )

    assert prompt_calls == [
        {
            'items': [
                '2026-01-02 03:04:05  SELECT 1 FROM dual',
                '2026-01-03 12:00:00  SELECT 2',
            ],
            'options': '--info=hidden --scheme=history --tiebreak=index --bind=ctrl-r:up,alt-r:up '
            '--preview-window=down:wrap:nohidden --no-height '
            "--preview=\"printf '%s' {} | pygmentize -l mysql -P style='monokai style'\"",
        }
    ]
    assert invalidated_apps == [event.app]
    assert event.current_buffer.text == 'SELECT  1\nFROM dual'
    assert event.current_buffer.cursor_position == len('SELECT  1\nFROM dual')


@pytest.mark.parametrize(
    ("highlight_preview", "pygmentize_available"),
    [
        (False, False),
        (False, True),
        (True, False),
    ],
)
def test_search_history_without_result_keeps_buffer_and_uses_plain_preview(
    monkeypatch,
    highlight_preview: bool,
    pygmentize_available: bool,
) -> None:
    prompt_calls: list[dict[str, Any]] = []
    invalidated_apps: list[Any] = []

    event = make_event(history=DummyHistory([('SELECT 1', '2026-01-01 00:00:00')]))

    class PromptingFzf:
        def is_available(self) -> bool:
            return True

        def prompt(self, items: list[str], fzf_options: str) -> list[str]:
            prompt_calls.append({'items': items, 'options': fzf_options})
            return []

    monkeypatch.setattr(fzf_module, 'Fzf', PromptingFzf)
    monkeypatch.setattr(
        fzf_module,
        'which',
        lambda executable: '/usr/bin/pygmentize' if pygmentize_available and executable == 'pygmentize' else None,
    )
    monkeypatch.setattr(fzf_module, 'safe_invalidate_display', lambda app: invalidated_apps.append(app))

    fzf_module.search_history(cast(Any, event), highlight_preview=highlight_preview)

    assert prompt_calls == [
        {
            'items': ['2026-01-01 00:00:00  SELECT 1'],
            'options': '--info=hidden --scheme=history --tiebreak=index --bind=ctrl-r:up,alt-r:up '
            "--preview-window=down:wrap:nohidden --no-height --preview=\"printf '%s' {}\"",
        }
    ]
    assert invalidated_apps == [event.app]
    assert event.current_buffer.text == 'original'
    assert event.current_buffer.cursor_position == 0



import sys

import click

try:
    import paramiko
except ImportError:
    from mycli.packages.paramiko_stub import paramiko  # type: ignore[no-redef]


# it isn't cool that this utility function can exit(), but it is slated to be removed anyway
def read_ssh_config(ssh_config_path: str):
    ssh_config = paramiko.config.SSHConfig()
    try:
        with open(ssh_config_path) as f:
            ssh_config.parse(f)
    except FileNotFoundError as e:
        click.secho(str(e), err=True, fg="red")
        sys.exit(1)
    # Paramiko prior to version 2.7 raises Exception on parse errors.
    # In 2.7 it has become paramiko.ssh_exception.SSHException,
    # but let's catch everything for compatibility
    except Exception as err:
        click.secho(f"Could not parse SSH configuration file {ssh_config_path}:\n{err} ", err=True, fg="red")
        sys.exit(1)
    else:
        return ssh_config



import os
from typing import Union

from prompt_toolkit.history import FileHistory

from mycli.packages.sql_utils import is_password_change

_StrOrBytesPath = Union[str, bytes, os.PathLike]


class FileHistoryWithTimestamp(FileHistory):
    """
    :class:`.FileHistory` class that stores all strings in a file with timestamp.
    """

    def __init__(self, filename: _StrOrBytesPath) -> None:
        self.filename = filename
        super().__init__(filename)

    def append_string(self, string: str) -> None:
        "Add string to the history."
        self._loaded_strings.insert(0, string)
        if is_password_change(string):
            return
        self.store_string(string)

    def load_history_with_timestamp(self) -> list[tuple[str, str]]:
        """
        Load history entries along with their timestamps.

        Returns:
            list[tuple[str, str]]: A list of tuples where each tuple contains
                                   a history entry and its corresponding timestamp.
        """
        history_with_timestamp: list[tuple[str, str]] = []
        lines: list[str] = []
        timestamp: str = ""

        def add() -> None:
            if lines:
                # Join and drop trailing newline.
                string = "".join(lines)[:-1]
                history_with_timestamp.append((string, timestamp))

        if os.path.exists(self.filename):
            with open(self.filename, 'r', encoding='utf-8') as f:
                for line in f:
                    if line.startswith("#"):
                        # Extract timestamp
                        timestamp = line[2:].strip()
                    elif line.startswith("+"):
                        lines.append(line[1:])
                    else:
                        add()
                        lines = []

                add()

        return list(reversed(history_with_timestamp))



from __future__ import annotations

import os
import sys
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Callable
from urllib.parse import parse_qs, unquote, urlparse

import click

from mycli.config import str_to_bool
from mycli.constants import EMPTY_PASSWORD_FLAG_SENTINEL, ISSUES_URL, REPO_URL
from mycli.main_modes.batch import main_batch_from_stdin, main_batch_with_progress_bar, main_batch_without_progress_bar
from mycli.main_modes.checkup import main_checkup
from mycli.main_modes.execute import main_execute_from_cli
from mycli.main_modes.list_dsn import main_list_dsn
from mycli.main_modes.list_ssh_config import main_list_ssh_config
from mycli.packages.cli_utils import is_valid_connection_scheme
from mycli.packages.ssh_utils import read_ssh_config

if TYPE_CHECKING:
    from mycli.main import CliArgs

ClientFactory = Callable[..., Any]


def run_from_cli_args(cli_args: 'CliArgs', client_factory: ClientFactory) -> None:
    from mycli import main as main_module

    cli_verbosity = main_module.preprocess_cli_args(cli_args, is_valid_connection_scheme)

    mycli = client_factory(
        prompt=cli_args.prompt,
        toolbar_format=cli_args.toolbar,
        logfile=cli_args.logfile,
        defaults_suffix=cli_args.defaults_group_suffix,
        defaults_file=cli_args.defaults_file,
        login_path=cli_args.login_path,
        auto_vertical_output=cli_args.auto_vertical_output,
        warn=cli_args.warn,
        myclirc=cli_args.myclirc,
        show_warnings=cli_args.show_warnings,
        cli_verbosity=cli_verbosity,
    )

    if cli_args.checkup:
        main_checkup(mycli)
        sys.exit(0)

    if cli_args.csv and cli_args.format not in [None, 'csv']:
        click.secho("Conflicting --csv and --format arguments.", err=True, fg="red")
        sys.exit(1)

    if cli_args.table and cli_args.format not in [None, 'table']:
        click.secho("Conflicting --table and --format arguments.", err=True, fg="red")
        sys.exit(1)

    if not cli_args.format:
        cli_args.format = 'default'

    if cli_args.csv:
        cli_args.format = 'csv'

    if cli_args.table:
        cli_args.format = 'table'

    if cli_args.deprecated_ssl is not None:
        click.secho(
            "Warning: The --ssl/--no-ssl CLI options are deprecated and will be removed in a future release. "
            "Please use the \"default_ssl_mode\" config option or --ssl-mode CLI flag instead. "
            f"See issue {ISSUES_URL}/1507",
            err=True,
            fg="yellow",
        )

    # ssh_port and ssh_config_path have truthy defaults and are not included
    if (
        any([
            cli_args.ssh_user,
            cli_args.ssh_host,
            cli_args.ssh_password,
            cli_args.ssh_key_filename,
            cli_args.list_ssh_config,
            cli_args.ssh_config_host,
        ])
        and not cli_args.ssh_warning_off
    ):
        click.secho(
            f"Warning: The built-in SSH functionality is deprecated and will be removed in a future release. See issue {ISSUES_URL}/1464",
            err=True,
            fg="red",
        )

    if cli_args.list_dsn:
        sys.exit(main_list_dsn(mycli))

    if cli_args.list_ssh_config:
        sys.exit(main_list_ssh_config(mycli, cli_args))

    if 'MYSQL_UNIX_PORT' in os.environ:
        # deprecated 2026-03
        click.secho(
            "The MYSQL_UNIX_PORT environment variable is deprecated in favor of MYSQL_UNIX_SOCKET.  "
            "MYSQL_UNIX_PORT will be removed in a future release.",
            err=True,
            fg="red",
        )
        if not cli_args.socket:
            cli_args.socket = os.environ['MYSQL_UNIX_PORT']

    if 'DSN' in os.environ:
        # deprecated 2026-03
        click.secho(
            "The DSN environment variable is deprecated in favor of MYSQL_DSN.  Support for DSN will be removed in a future release.",
            err=True,
            fg="red",
        )
        if not cli_args.dsn:
            cli_args.dsn = os.environ['DSN']

    # Choose which ever one has a valid value.
    database = cli_args.dbname or cli_args.database

    dsn_uri = None

    # Treat the database argument as a DSN alias only if it matches a configured alias
    # todo why is port tested but not socket?
    truthy_password = cli_args.password not in (None, EMPTY_PASSWORD_FLAG_SENTINEL)
    if (
        database
        and "://" not in database
        and not any([
            cli_args.user,
            truthy_password,
            cli_args.host,
            cli_args.port,
            cli_args.login_path,
        ])
        and database in mycli.config.get("alias_dsn", {})
    ):
        cli_args.dsn, database = database, ""

    if database and "://" in database:
        dsn_uri, database = database, ""

    if cli_args.dsn:
        try:
            dsn_uri = mycli.config["alias_dsn"][cli_args.dsn]
        except KeyError:
            is_valid_scheme, scheme = is_valid_connection_scheme(cli_args.dsn)
            if is_valid_scheme:
                dsn_uri = cli_args.dsn
            else:
                click.secho(
                    "Could not find the specified DSN in the config file. Please check the \"[alias_dsn]\" section in your myclirc.",
                    err=True,
                    fg="red",
                )
                sys.exit(1)
        else:
            mycli.dsn_alias = cli_args.dsn

    if dsn_uri:
        uri = urlparse(dsn_uri)
        if not database:
            database = uri.path[1:]  # ignore the leading fwd slash
        if not cli_args.user and uri.username is not None:
            cli_args.user = unquote(uri.username)
        # todo: rationalize the behavior of empty-string passwords here
        if not cli_args.password and uri.password is not None:
            cli_args.password = unquote(uri.password)
        if not cli_args.host:
            cli_args.host = uri.hostname
        if not cli_args.port:
            cli_args.port = uri.port

        if uri.query:
            dsn_params = parse_qs(uri.query)
        else:
            dsn_params = {}

        if params := dsn_params.get('ssl'):
            click.secho(
                'Warning: The "ssl" DSN URI parameter is deprecated and will be removed in a future release. '
                'Please use the "ssl_mode" parameter instead. '
                f'See issue {ISSUES_URL}/1507',
                err=True,
                fg='yellow',
            )
            if params[0].lower() == 'true':
                cli_args.ssl_mode = 'on'
        if params := dsn_params.get('ssl_mode'):
            cli_args.ssl_mode = cli_args.ssl_mode or params[0]
        if params := dsn_params.get('ssl_ca'):
            cli_args.ssl_ca = cli_args.ssl_ca or params[0]
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('ssl_capath'):
            cli_args.ssl_capath = cli_args.ssl_capath or params[0]
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('ssl_cert'):
            cli_args.ssl_cert = cli_args.ssl_cert or params[0]
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('ssl_key'):
            cli_args.ssl_key = cli_args.ssl_key or params[0]
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('ssl_cipher'):
            cli_args.ssl_cipher = cli_args.ssl_cipher or params[0]
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('tls_version'):
            cli_args.tls_version = cli_args.tls_version or params[0]
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('ssl_verify_server_cert'):
            cli_args.ssl_verify_server_cert = cli_args.ssl_verify_server_cert or (params[0].lower() == 'true')
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('socket'):
            cli_args.socket = cli_args.socket or params[0]
        if params := dsn_params.get('keepalive_ticks'):
            if cli_args.keepalive_ticks is None:
                cli_args.keepalive_ticks = int(params[0])
        if params := dsn_params.get('character_set'):
            cli_args.character_set = cli_args.character_set or params[0]

    keepalive_ticks = cli_args.keepalive_ticks if cli_args.keepalive_ticks is not None else mycli.default_keepalive_ticks
    ssl_mode = cli_args.ssl_mode or mycli.ssl_mode

    # if there is a mismatch between the ssl_mode value and other sources of ssl config, show a warning
    # specifically using "is False" to not pickup the case where cli_args.deprecated_ssl is None (not set by the user)
    if cli_args.deprecated_ssl and ssl_mode == "off" or cli_args.deprecated_ssl is False and ssl_mode in ("auto", "on"):
        click.secho(
            f"Warning: The current ssl_mode value of '{ssl_mode}' is overriding the value provided by "
            f"either the --ssl/--no-ssl CLI options or a DSN URI parameter (ssl={cli_args.deprecated_ssl}).",
            err=True,
            fg="yellow",
        )

    # configure SSL if ssl_mode is auto/on or if
    # cli_args.deprecated_ssl = True (from --ssl or a DSN URI) and ssl_mode is None
    if ssl_mode in ("auto", "on") or (cli_args.deprecated_ssl and ssl_mode is None):
        if cli_args.socket and ssl_mode == 'auto':
            ssl = None
        else:
            ssl = {
                "mode": ssl_mode,
                "enable": cli_args.deprecated_ssl,  # todo: why is this set at all?
                "ca": cli_args.ssl_ca and os.path.expanduser(cli_args.ssl_ca),
                "cert": cli_args.ssl_cert and os.path.expanduser(cli_args.ssl_cert),
                "key": cli_args.ssl_key and os.path.expanduser(cli_args.ssl_key),
                "capath": cli_args.ssl_capath,
                "cipher": cli_args.ssl_cipher,
                "tls_version": cli_args.tls_version,
                "check_hostname": cli_args.ssl_verify_server_cert,
            }
            # remove empty ssl options
            ssl = {k: v for k, v in ssl.items() if v is not None}
    else:
        ssl = None

    if cli_args.ssh_config_host:
        ssh_config = read_ssh_config(cli_args.ssh_config_path).lookup(cli_args.ssh_config_host)
        ssh_host = cli_args.ssh_host if cli_args.ssh_host else ssh_config.get("hostname")
        ssh_user = cli_args.ssh_user if cli_args.ssh_user else ssh_config.get("user")
        if ssh_config.get("port") and cli_args.ssh_port == 22:
            # port has a default value, overwrite it if it's in the config
            ssh_port = int(ssh_config.get("port"))
        else:
            ssh_port = cli_args.ssh_port
        ssh_key_filename = cli_args.ssh_key_filename if cli_args.ssh_key_filename else ssh_config.get("identityfile", [None])[0]
    else:
        ssh_host = cli_args.ssh_host
        ssh_user = cli_args.ssh_user
        ssh_port = cli_args.ssh_port
        ssh_key_filename = cli_args.ssh_key_filename

    ssh_key_filename = ssh_key_filename and os.path.expanduser(ssh_key_filename)
    # Merge init-commands: global, DSN-specific, then CLI
    init_cmds: list[str] = []
    # 1) Global init-commands
    global_section = mycli.config.get("init-commands", {})
    for _, val in global_section.items():
        if isinstance(val, (list, tuple)):
            init_cmds.extend(val)
        elif val:
            init_cmds.append(val)
    # 2) DSN-specific init-commands
    if cli_args.dsn:
        alias_section = mycli.config.get("alias_dsn.init-commands", {})
        if cli_args.dsn in alias_section:
            val = alias_section.get(cli_args.dsn)
            if isinstance(val, (list, tuple)):
                init_cmds.extend(val)
            elif val:
                init_cmds.append(val)
    # 3) CLI-provided init_command
    if cli_args.init_command:
        init_cmds.append(cli_args.init_command)

    combined_init_cmd = "; ".join(cmd.strip() for cmd in init_cmds if cmd)

    if cli_args.use_keyring is not None and cli_args.use_keyring.lower() == 'reset':
        use_keyring = True
        reset_keyring = True
    elif cli_args.use_keyring is None:
        use_keyring = str_to_bool(mycli.config['main'].get('use_keyring', 'False'))
        reset_keyring = False
    else:
        use_keyring = str_to_bool(cli_args.use_keyring)
        reset_keyring = False

    # todo: removeme after a period of transition
    for tup in [
        ('client', 'prompt', 'prompt', 'main', 'prompt'),
        ('client', 'pager', 'pager', 'main', 'pager'),
        ('client', 'skip-pager', 'skip-pager', 'main', 'enable_pager'),
        # this is a white lie, because default_character_set can actually be read from the package config
        ('client', 'default-character-set', 'default-character-set', 'connection', 'default_character_set'),
        # local-infile can be read from both sections
        ('mysqld', 'local-infile', 'local-infile', 'connection', 'default_local_infile'),
        ('client', 'local-infile', 'local-infile', 'connection', 'default_local_infile'),
        ('mysqld', 'loose-local-infile', 'loose-local-infile', 'connection', 'default_local_infile'),
        ('client', 'loose-local-infile', 'loose-local-infile', 'connection', 'default_local_infile'),
        # todo: in the future we should add default_port, etc, but only in .myclirc
        # they are currently ignored in my.cnf
        ('mysqld', 'default_socket', 'socket', 'connection', 'default_socket'),
        ('client', 'ssl-ca', 'ssl-ca', 'connection', 'default_ssl_ca'),
        ('client', 'ssl-cert', 'ssl-cert', 'connection', 'default_ssl_cert'),
        ('client', 'ssl-key', 'ssl-key', 'connection', 'default_ssl_key'),
        ('client', 'ssl-cipher', 'ssl-cipher', 'connection', 'default_ssl_cipher'),
        ('client', 'ssl-verify-server-cert', 'ssl-verify-server-cert', 'connection', 'default_ssl_verify_server_cert'),
    ]:
        (
            mycnf_section_name,
            mycnf_item_name,
            printable_mycnf_item_name,
            myclirc_section_name,
            myclirc_item_name,
        ) = tup
        if str_to_bool(mycli.config['main'].get('my_cnf_transition_done', 'False')):
            break
        if (
            mycli.my_cnf[mycnf_section_name].get(mycnf_item_name) is None
            and mycli.my_cnf[mycnf_section_name].get(mycnf_item_name.replace('-', '_')) is None
        ):
            continue
        user_section = mycli.config_without_package_defaults.get(myclirc_section_name, {})
        if user_section.get(myclirc_item_name) is None:
            cnf_value = mycli.my_cnf[mycnf_section_name].get(mycnf_item_name)
            if cnf_value is None:
                cnf_value = mycli.my_cnf[mycnf_section_name].get(mycnf_item_name.replace('-', '_'))
            click.secho(
                dedent(
                    f"""
                    Reading configuration from my.cnf files is deprecated.
                    See {ISSUES_URL}/1490 .
                    The cause of this message is the following in a my.cnf file without a corresponding
                    ~/.myclirc entry:

                        [{mycnf_section_name}]
                        {printable_mycnf_item_name} = {cnf_value}

                    To suppress this message, remove the my.cnf item add or the following to ~/.myclirc:

                        [{myclirc_section_name}]
                        {myclirc_item_name} = <value>

                    The ~/.myclirc setting will take precedence.  In the future, the my.cnf will be ignored.

                    Values are documented at {REPO_URL}/blob/main/mycli/myclirc .  An
                    empty <value> is generally accepted.

                    To ignore all of this, set

                        [main]
                        my_cnf_transition_done = True

                    in ~/.myclirc.

                    --------

                    """
                ),
                err=True,
                fg='yellow',
            )

    mycli.connect(
        database=database,
        user=cli_args.user,
        passwd=cli_args.password,
        host=cli_args.host,
        port=cli_args.port,
        socket=cli_args.socket,
        local_infile=cli_args.local_infile,
        ssl=ssl,
        ssh_user=ssh_user,
        ssh_host=ssh_host,
        ssh_port=ssh_port,
        ssh_password=cli_args.ssh_password,
        ssh_key_filename=ssh_key_filename,
        init_command=combined_init_cmd,
        unbuffered=cli_args.unbuffered,
        character_set=cli_args.character_set,
        use_keyring=use_keyring,
        reset_keyring=reset_keyring,
        keepalive_ticks=keepalive_ticks,
    )

    if combined_init_cmd:
        click.echo(f"Executing init-command: {combined_init_cmd}", err=True)

    mycli.logger.debug(
        "Launch Params: \n\tdatabase: %r\tuser: %r\thost: %r\tport: %r",
        database,
        cli_args.user,
        cli_args.host,
        cli_args.port,
    )

    if cli_args.execute is not None:
        sys.exit(main_execute_from_cli(mycli, cli_args))

    if cli_args.batch is not None and cli_args.batch != '-' and cli_args.progress and sys.stderr.isatty():
        sys.exit(main_batch_with_progress_bar(mycli, cli_args))

    if cli_args.batch is not None:
        sys.exit(main_batch_without_progress_bar(mycli, cli_args))

    if not sys.stdin.isatty():
        sys.exit(main_batch_from_stdin(mycli, cli_args))

    mycli.run_cli()
    mycli.close()



import importlib.metadata
import json
import os
import shutil
import sys
import urllib.error
import urllib.request

from mycli.constants import REPO_URL

PYPI_API_BASE = 'https://pypi.org/pypi'


def pypi_api_fetch(fragment: str) -> dict:
    fragment = fragment.lstrip('/')
    url = f'{PYPI_API_BASE}/{fragment}'
    try:
        with urllib.request.urlopen(url, timeout=5) as response:
            return json.loads(response.read().decode('utf8'))
    except urllib.error.URLError:
        print(f'Failed to connect to PyPi on {url}', file=sys.stderr)
        return {}


def _dependencies_checkup() -> None:
    print('\n### Key Python dependencies:\n')
    for dependency in [
        'cli_helpers',
        'click',
        'prompt_toolkit',
        'pygments',
        'pymysql',
        'sqlglot',
        'sqlglotc',
        'tabulate',
    ]:
        try:
            installed_version = importlib.metadata.version(dependency)
        except importlib.metadata.PackageNotFoundError:
            installed_version = None
        pypi_profile = pypi_api_fetch(f'/{dependency}/json')
        latest_version = pypi_profile.get('info', {}).get('version', None)
        print(f'{dependency} version {installed_version} (latest {latest_version})')


def _executables_checkup() -> None:
    print('\n### External executables:\n')
    for executable in [
        'less',
        'fzf',
        'pygmentize',
    ]:
        if shutil.which(executable):
            print(f'The "{executable}" executable was found — good!')
        else:
            print(f'The recommended "{executable}" executable was not found — some functionality will suffer.')


def _environment_checkup() -> None:
    print('\n### Environment variables:\n')
    for variable in [
        'EDITOR',
        'VISUAL',
    ]:
        if value := os.environ.get(variable):
            print(f'The ${variable} environment variable was set to "{value}" — good!')
        else:
            print(f'The ${variable} environment variable was not set — some functionality will suffer.')


def _configuration_checkup(mycli) -> None:
    did_output_missing = False
    did_output_unsupported = False
    did_output_deprecated = False

    indent = '    '
    transitions = {
        f'{indent}[main]\n{indent}default_character_set': f'{indent}[connection]\n{indent}default_character_set',
        f'{indent}[main]\n{indent}ssl_mode': f'{indent}[connection]\n{indent}default_ssl_mode',
    }
    reverse_transitions = {v: k for k, v in transitions.items()}

    if not list(mycli.config.keys()):
        print('\n### Missing file:\n')
        print('The local ~/,myclirc is missing or empty.\n')
        did_output_missing = True
    else:
        for section_name in mycli.config:
            if section_name not in mycli.config_without_package_defaults:
                if not did_output_missing:
                    print('\n### Missing in user ~/.myclirc:\n')
                print(f'The entire section:\n\n{indent}[{section_name}]\n')
                did_output_missing = True
                continue
            for item_name in mycli.config[section_name]:
                transition_key = f'{indent}[{section_name}]\n{indent}{item_name}'
                if transition_key in reverse_transitions:
                    continue
                if item_name not in mycli.config_without_package_defaults[section_name]:
                    if not did_output_missing:
                        print('\n### Missing in user ~/.myclirc:\n')
                    print(f'The item:\n\n{indent}[{section_name}]\n{indent}{item_name} =\n')
                    did_output_missing = True

        for section_name in mycli.config_without_package_defaults:
            if section_name not in mycli.config_without_user_options:
                if not did_output_unsupported:
                    print('\n### Unsupported in user ~/.myclirc:\n')
                did_output_unsupported = True
                print(f'The entire section:\n\n{indent}[{section_name}]\n')
                continue
            for item_name in mycli.config_without_package_defaults[section_name]:
                if section_name == 'colors' and item_name.startswith('sql.'):
                    # these are commented out in the package myclirc
                    continue
                if section_name in [
                    'favorite_queries',
                    'init-commands',
                    'alias_dsn',
                    'alias_dsn.init-commands',
                ]:
                    # these are free-entry sections, so a comparison per item is not meaningful
                    continue
                transition_key = f'{indent}[{section_name}]\n{indent}{item_name}'
                if transition_key in transitions:
                    continue
                if item_name not in mycli.config_without_user_options[section_name]:
                    if not did_output_unsupported:
                        print('\n### Unsupported in user ~/.myclirc:\n')
                    print(f'The item:\n\n{indent}[{section_name}]\n{indent}{item_name} =\n')
                    did_output_unsupported = True

        for section_name in mycli.config_without_package_defaults:
            if section_name not in mycli.config_without_user_options:
                continue
            for item_name in mycli.config_without_package_defaults[section_name]:
                if section_name == 'colors' and item_name.startswith('sql.'):
                    # these are commented out in the package myclirc
                    continue
                transition_key = f'{indent}[{section_name}]\n{indent}{item_name}'
                if transition_key in transitions:
                    if not did_output_deprecated:
                        print('\n### Deprecated in user ~/.myclirc:\n')
                    transition_value = transitions[transition_key]
                    print(f'It is recommended to transition:\n\n{transition_key}\n\nto\n\n{transition_value}\n')
                    did_output_deprecated = True

    if did_output_missing or did_output_unsupported or did_output_deprecated:
        print(f'For more info on supported features, see the commentary and defaults at:\n\n    * {REPO_URL}/blob/main/mycli/myclirc\n')
    else:
        print('\n### Configuration:\n')
        print('User configuration all up to date!\n')


def main_checkup(mycli) -> None:
    _dependencies_checkup()
    _executables_checkup()
    _environment_checkup()
    _configuration_checkup(mycli)



from __future__ import annotations

from io import TextIOWrapper
import logging
import os
import threading
from typing import IO, Literal

from cli_helpers.tabular_output import TabularOutputFormatter
from prompt_toolkit.formatted_text import to_formatted_text
from prompt_toolkit.shortcuts import PromptSession
import sqlparse

from mycli.app_state import (
    AppStateMixin,
    configure_prompt_state,
    destructive_keywords_from_config,
    ensure_my_cnf_sections,
    llm_prompt_truncation,
    normalize_ssl_mode,
)
from mycli.client_commands import ClientCommandsMixin
from mycli.client_connection import ClientConnectionMixin
from mycli.client_query import ClientQueryMixin
from mycli.clistyle import style_factory_helpers, style_factory_ptoolkit
from mycli.completion_refresher import CompletionRefresher
from mycli.config import get_mylogin_cnf_path, open_mylogin_cnf, read_config_files, write_default_config
from mycli.constants import DEFAULT_PROMPT
from mycli.main_modes import repl as repl_package
from mycli.output import OutputMixin
from mycli.packages import special
from mycli.packages.special.favoritequeries import FavoriteQueries
from mycli.packages.tabular_output import sql_format
from mycli.schema_prefetcher import SchemaPrefetcher
from mycli.sqlcompleter import SQLCompleter
from mycli.sqlexecute import SQLExecute
from mycli.types import Query

sqlparse.engine.grouping.MAX_GROUPING_DEPTH = None  # type: ignore[assignment]
sqlparse.engine.grouping.MAX_GROUPING_TOKENS = None  # type: ignore[assignment]


class MyCli(AppStateMixin, OutputMixin, ClientCommandsMixin, ClientConnectionMixin, ClientQueryMixin):
    default_prompt = DEFAULT_PROMPT
    default_prompt_splitln = "\\u@\\h\\n(\\t):\\d>"
    max_len_prompt = 45
    defaults_suffix = None
    prompt_lines: int
    sqlexecute: SQLExecute | None
    numeric_alignment: str

    # In order of being loaded. Files lower in list override earlier ones.
    cnf_files: list[str | IO[str]] = [
        "/etc/my.cnf",
        "/etc/mysql/my.cnf",
        "/usr/local/etc/my.cnf",
        os.path.expanduser("~/.my.cnf"),
    ]

    # check XDG_CONFIG_HOME exists and not an empty string
    xdg_config_home = os.environ.get("XDG_CONFIG_HOME", "~/.config")
    system_config_files: list[str | IO[str]] = [
        "/etc/myclirc",
        os.path.join(os.path.expanduser(xdg_config_home), "mycli", "myclirc"),
    ]

    pwd_config_file = os.path.join(os.getcwd(), ".myclirc")

    def __init__(
        self,
        sqlexecute: SQLExecute | None = None,
        prompt: str | None = None,
        toolbar_format: str | None = None,
        logfile: TextIOWrapper | Literal[False] | None = None,
        defaults_suffix: str | None = None,
        defaults_file: str | None = None,
        login_path: str | None = None,
        auto_vertical_output: bool = False,
        warn: bool | None = None,
        myclirc: str = "~/.myclirc",
        show_warnings: bool | None = None,
        cli_verbosity: int = 0,
    ) -> None:
        self.sqlexecute = sqlexecute
        self.logfile = logfile
        self.defaults_suffix = defaults_suffix
        self.login_path = login_path
        self.toolbar_error_message: str | None = None
        self.prompt_session: PromptSession | None = None
        self._keepalive_counter = 0
        self.keepalive_ticks: int | None = 0
        self.sandbox_mode: bool = False

        # self.cnf_files is a class variable that stores the list of mysql
        # config files to read in at launch.
        # If defaults_file is specified then override the class variable with
        # defaults_file.
        if defaults_file:
            self.cnf_files = [defaults_file]

        # Load config.
        config_files: list[str | IO[str]] = self.system_config_files + [myclirc] + [self.pwd_config_file]

        c = self.config = read_config_files(config_files)
        # this parallel config exists to
        #  * compare with my.cnf
        #  * support the --checkup feature
        # todo: after removing my.cnf, create the parallel configs only when --checkup is set
        self.config_without_package_defaults = read_config_files(config_files, ignore_package_defaults=True)
        # this parallel config exists to compare with my.cnf support the --checkup feature
        self.config_without_user_options = read_config_files(config_files, ignore_user_options=True)
        self.multi_line = c["main"].as_bool("multi_line")
        self.key_bindings = c["main"]["key_bindings"]
        self.emacs_ttimeoutlen = c['keys'].as_float('emacs_ttimeoutlen')
        self.vi_ttimeoutlen = c['keys'].as_float('vi_ttimeoutlen')
        special.set_timing_enabled(c["main"].as_bool("timing"))
        special.set_show_favorite_query(c["main"].as_bool("show_favorite_query"))
        if show_warnings is not None:
            special.set_show_warnings_enabled(show_warnings)
        else:
            special.set_show_warnings_enabled(c['main'].as_bool('show_warnings'))
        self.beep_after_seconds = float(c["main"]["beep_after_seconds"] or 0)
        self.default_keepalive_ticks = c['connection'].as_int('default_keepalive_ticks')

        FavoriteQueries.instance = FavoriteQueries.from_config(self.config)

        self.dsn_alias: str | None = None
        self.main_formatter = TabularOutputFormatter(format_name=c["main"]["table_format"])
        self.redirect_formatter = TabularOutputFormatter(format_name=c["main"].get("redirect_format", "csv"))
        sql_format.register_new_formatter(self.main_formatter)
        sql_format.register_new_formatter(self.redirect_formatter)
        self.main_formatter.mycli = self
        self.redirect_formatter.mycli = self
        self.syntax_style = c["main"]["syntax_style"]
        self.verbosity = -1 if c["main"].as_bool("less_chatty") else 0
        if cli_verbosity:
            self.verbosity = cli_verbosity
        self.cli_style = c["colors"]
        self.ptoolkit_style = style_factory_ptoolkit(self.syntax_style, self.cli_style)
        self.helpers_style = style_factory_helpers(self.syntax_style, self.cli_style)
        self.helpers_warnings_style = style_factory_helpers(self.syntax_style, self.cli_style, warnings=True)
        self.wider_completion_menu = c["main"].as_bool("wider_completion_menu")
        c_dest_warning = c["main"].as_bool("destructive_warning")
        self.destructive_warning = c_dest_warning if warn is None else warn
        self.login_path_as_host = c["main"].as_bool("login_path_as_host")
        self.post_redirect_command = c['main'].get('post_redirect_command')
        self.null_string = c['main'].get('null_string')
        self.numeric_alignment = c['main'].get('numeric_alignment', 'right') or 'right'
        self.binary_display = c['main'].get('binary_display')
        self.llm_prompt_field_truncate, self.llm_prompt_section_truncate = llm_prompt_truncation(c)

        self.ssl_mode, ssl_mode_error = normalize_ssl_mode(c)
        if ssl_mode_error:
            self.echo(ssl_mode_error, err=True, fg="red")

        # read from cli argument or user config file
        self.auto_vertical_output = auto_vertical_output or c["main"].as_bool("auto_vertical_output")

        # Write user config if system config wasn't the last config loaded.
        if c.filename not in self.system_config_files and not os.path.exists(myclirc):
            write_default_config(myclirc)

        # audit log
        if self.logfile is None and "audit_log" in c["main"]:
            try:
                self.logfile = open(os.path.expanduser(c["main"]["audit_log"]), "a")
            except (IOError, OSError):
                self.echo("Error: Unable to open the audit log file. Your queries will not be logged.", err=True, fg="red")
                self.logfile = False

        self.completion_refresher = CompletionRefresher()
        self.prefetch_schemas_mode = c["main"].get("prefetch_schemas_mode", "always") or "always"
        raw_prefetch_list = c["main"].as_list("prefetch_schemas_list") if "prefetch_schemas_list" in c["main"] else []
        self.prefetch_schemas_list = [s.strip() for s in raw_prefetch_list if s and s.strip()]
        self.schema_prefetcher = SchemaPrefetcher(self)

        self.logger = logging.getLogger(__name__)
        self.initialize_logging()

        keyword_casing = c["main"].get("keyword_casing", "auto")

        self.highlight_preview = c['search'].as_bool('highlight_preview')

        self.query_history: list[Query] = []

        # Initialize completer.
        self.smart_completion = c["main"].as_bool("smart_completion")
        self.completer = SQLCompleter(
            self.smart_completion, supported_formats=self.main_formatter.supported_formats, keyword_casing=keyword_casing
        )
        self._completer_lock = threading.Lock()

        self.min_completion_trigger = c["main"].as_int("min_completion_trigger")
        # a hack, pending a better way to handle settings and state
        repl_package.MIN_COMPLETION_TRIGGER = self.min_completion_trigger
        self.last_prompt_message = to_formatted_text('')
        self.last_custom_toolbar_message = to_formatted_text('')

        # Register custom special commands.
        self.register_special_commands()

        # Load .mylogin.cnf if it exists.
        mylogin_cnf_path = get_mylogin_cnf_path()
        if mylogin_cnf_path:
            mylogin_cnf = open_mylogin_cnf(mylogin_cnf_path)
            if mylogin_cnf_path and mylogin_cnf:
                # .mylogin.cnf gets read last, even if defaults_file is specified.
                self.cnf_files.append(mylogin_cnf)
            elif mylogin_cnf_path and not mylogin_cnf:
                # There was an error reading the login path file.
                print("Error: Unable to read login path file.")

        self.my_cnf = read_config_files(self.cnf_files, list_values=False)
        ensure_my_cnf_sections(self.my_cnf)
        prompt_cnf = self.read_my_cnf(self.my_cnf, ["prompt"])["prompt"]
        configure_prompt_state(self, c, prompt, prompt_cnf, toolbar_format)
        self.prompt_session = None
        self.destructive_keywords = destructive_keywords_from_config(c)
        special.set_destructive_keywords(self.destructive_keywords)

    def close(self) -> None:
        if hasattr(self, 'schema_prefetcher'):
            self.schema_prefetcher.stop()
        if self.sqlexecute is not None:
            self.sqlexecute.close()

    def run_cli(self) -> None:
        repl_package.main_repl(self)
